# coding: utf-8

# Copyright 2024 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re


# This is the name of the primary logger used by the library.
LOGGER_NAME = 'ibm-cloud-sdk-core'
# Keywords that are redacted.
REDACTED_KEYWORDS = [
    "apikey",
    "api_key",
    "passcode",
    "password",
    "token",
    "aadClientId",
    "aadClientSecret",
    "auth",
    "auth_provider_x509_cert_url",
    "auth_uri",
    "client_email",
    "client_id",
    "client_x509_cert_url",
    "key",
    "project_id",
    "secret",
    "subscriptionId",
    "tenantId",
    "thumbprint",
    "token_uri",
]


class LoggingFilter:
    """Functions used to filter messages before they are logged."""

    redacted_tokens = "|".join(REDACTED_KEYWORDS)
    auth_header_pattern = re.compile(r"(?m)(Authorization|X-Auth\S*): ((.*?)(\r\n.*)|(.*))")
    property_settings_pattern = re.compile(r"(?i)(" + redacted_tokens + r")=[^&]*(&|$)")
    json_field_pattern = re.compile(r'(?i)"([^"]*(' + redacted_tokens + r')[^"_]*)":\s*"[^\,]*"')

    @classmethod
    def redact_secrets(cls, text: str) -> str:
        """Replaces values of potential secret keywords with a placeholder value.
        Args:
            text (str): the string to check and process

        Returns:
            str: the safe, redacted string with all secrets masked out
        """

        placeholder = "[redacted]"
        redacted = cls.auth_header_pattern.sub(r"\1: " + placeholder + r"\4", text)
        redacted = cls.property_settings_pattern.sub(r"\1=" + placeholder + r"\2", redacted)
        redacted = cls.json_field_pattern.sub(r'"\1":"' + placeholder + r'"', redacted)
        return redacted

    @classmethod
    def filter_message(cls, s: str) -> str:
        """Filters 's' prior to logging it as a debug message"""
        # Redact secrets
        s = LoggingFilter.redact_secrets(s)

        # Replace CRLF characters with an actual newline to make the message more readable.
        s = s.replace('\\r\\n', '\n')
        return s


def get_logger() -> logging.Logger:
    """Returns the primary logger object instance used by the library."""
    return logging.getLogger(LOGGER_NAME)



# pylint: disable=missing-docstring
import logging
import json
import time
import jwt
import pytest
import responses

from ibm_cloud_sdk_core.authenticators import MCSPAuthenticator, Authenticator
from .utils.logger_utils import setup_test_logger

setup_test_logger(logging.ERROR)


OPERATION_PATH = '/siusermgr/api/1.0/apikeys/token'
MOCK_URL = 'https://mcsp.ibm.com'


def test_mcsp_authenticator():
    authenticator = MCSPAuthenticator('my-api-key', MOCK_URL)
    assert authenticator is not None
    assert authenticator.authentication_type() == Authenticator.AUTHTYPE_MCSP
    assert authenticator.token_manager.url == MOCK_URL
    assert authenticator.token_manager.disable_ssl_verification is False
    assert authenticator.token_manager.headers == {
        'Accept': 'application/json',
        'Content-Type': 'application/json',
    }
    assert authenticator.token_manager.proxies is None

    authenticator.set_disable_ssl_verification(True)
    assert authenticator.token_manager.disable_ssl_verification is True

    with pytest.raises(TypeError) as err:
        authenticator.set_headers('dummy')
    assert str(err.value) == 'headers must be a dictionary'

    authenticator.set_headers({'dummy': 'headers'})
    assert authenticator.token_manager.headers == {'dummy': 'headers'}

    with pytest.raises(TypeError) as err:
        authenticator.set_proxies('dummy')
    assert str(err.value) == 'proxies must be a dictionary'

    authenticator.set_proxies({'dummy': 'proxies'})
    assert authenticator.token_manager.proxies == {'dummy': 'proxies'}


def test_disable_ssl_verification():
    authenticator = MCSPAuthenticator('my-api-key', MOCK_URL, disable_ssl_verification=True)
    assert authenticator.token_manager.disable_ssl_verification is True

    authenticator.set_disable_ssl_verification(False)
    assert authenticator.token_manager.disable_ssl_verification is False


def test_invalid_disable_ssl_verification_type():
    with pytest.raises(TypeError) as err:
        authenticator = MCSPAuthenticator('my-api-key', MOCK_URL, disable_ssl_verification='True')
    assert str(err.value) == 'disable_ssl_verification must be a bool'

    authenticator = MCSPAuthenticator('my-api-key', MOCK_URL)
    assert authenticator.token_manager.disable_ssl_verification is False

    with pytest.raises(TypeError) as err:
        authenticator.set_disable_ssl_verification('True')
    assert str(err.value) == 'status must be a bool'


def test_mcsp_authenticator_validate_failed():
    with pytest.raises(ValueError) as err:
        MCSPAuthenticator(apikey=None, url=MOCK_URL)
    assert str(err.value) == 'The apikey shouldn\'t be None.'

    with pytest.raises(ValueError) as err:
        MCSPAuthenticator(apikey='my-api-key', url=None)
    assert str(err.value) == 'The url shouldn\'t be None.'


# utility function to construct a mock token server response containing an access token.
def get_mock_token_response(issued_at, time_to_live) -> str:
    access_token_layout = {
        "username": "dummy",
        "role": "Admin",
        "permissions": ["administrator", "manage_catalog"],
        "sub": "admin",
        "iss": "sss",
        "aud": "sss",
        "uid": "sss",
        "iat": issued_at,
        "exp": issued_at + time_to_live,
    }

    access_token = jwt.encode(
        access_token_layout, 'secret', algorithm='HS256', headers={'kid': '230498151c214b788dd97f22b85410a5'}
    )

    token_server_response = {"token": access_token, "token_type": "jwt", "expires_in": time_to_live}
    # For convenience, return both the server response and the access_token.
    return (json.dumps(token_server_response), access_token)


@responses.activate
def test_get_token():
    (response, access_token) = get_mock_token_response(time.time(), 7200)
    responses.add(responses.POST, MOCK_URL + OPERATION_PATH, body=response, status=200)

    auth_headers = {'Host': 'mcsp.cloud.ibm.com:443'}
    authenticator = MCSPAuthenticator(apikey='my-api-key', url=MOCK_URL, headers=auth_headers)

    # Authenticate the request and verify the Authorization header.
    request = {'headers': {}}
    authenticator.authenticate(request)
    assert request['headers']['Authorization'] == 'Bearer ' + access_token

    # Verify that the "get token" request contained the Host header.
    assert responses.calls[0].request.headers.get('Host') == 'mcsp.cloud.ibm.com:443'


@responses.activate
def test_get_token_cached():
    (response, access_token) = get_mock_token_response(time.time(), 7200)
    responses.add(responses.POST, MOCK_URL + OPERATION_PATH, body=response, status=200)

    authenticator = MCSPAuthenticator(apikey='my-api-key', url=MOCK_URL)

    # Authenticate the request and verify the Authorization header.
    request = {'headers': {}}
    authenticator.authenticate(request)
    assert request['headers']['Authorization'] == 'Bearer ' + access_token

    # Authenticate a second request and verify that we used the same access token.
    request = {'headers': {}}
    authenticator.authenticate(request)
    assert request['headers']['Authorization'] == 'Bearer ' + access_token


@responses.activate
def test_get_token_background_refresh():
    t1 = time.time()
    t2 = t1 + 7200

    # Setup the first token response.
    (response1, access_token1) = get_mock_token_response(t1, 7200)
    responses.add(responses.POST, MOCK_URL + OPERATION_PATH, body=response1, status=200)

    # Setup the second token response.
    (response2, access_token2) = get_mock_token_response(t2, 7200)
    responses.add(responses.POST, MOCK_URL + OPERATION_PATH, body=response2, status=200)

    authenticator = MCSPAuthenticator(apikey="my-api-key", url=MOCK_URL)

    # Authenticate the request and verify that the first access_token is used.
    request = {'headers': {}}
    authenticator.authenticate(request)
    assert request['headers']['Authorization'] == 'Bearer ' + access_token1

    # Now put the token manager in the refresh window to trigger a background refresh scenario.
    authenticator.token_manager.refresh_time = t1 - 1

    # Authenticate a second request and verify that the correct access token is used.
    # Note: Ideally, the token manager would trigger the refresh in a separate thread
    # and it "should" return the first access token for this second authentication request
    # while the token manager is obtaining a new access token.
    # Unfortunately, the TokenManager class  method does the refresh request synchronously,
    # so we get back the second access token here instead.
    # If we "fix" the TokenManager class to refresh asynchronously, we'll need to
    # change this test case to expect the first access token here.
    request = {'headers': {}}
    authenticator.authenticate(request)
    assert request['headers']['Authorization'] == 'Bearer ' + access_token2

    # Wait for the background refresh to finish.
    # No need to wait due to the synchronous logic in the TokenManager class mentioned above.
    # time.sleep(2)

    # Authenticate another request and verify that the second access token is used again.
    request = {'headers': {}}
    authenticator.authenticate(request)
    assert request['headers']['Authorization'] == 'Bearer ' + access_token2



# coding: utf-8

# Copyright 2025. IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Dict, List, Optional
import requests

from ibm_cloud_sdk_core.logger import get_logger
from ..private_helpers import _build_user_agent
from .jwt_token_manager import JWTTokenManager

logger = get_logger()


class MCSPV2TokenManager(JWTTokenManager):
    """The MCSPV2TokenManager invokes the MCSP v2 token-exchange operation
    (POST /api/2.0/{scopeCollectionType}/{scopeId}/apikeys/token) to obtain an access token
    for an apikey. When the access token expires, a new access token is obtained from the token server.

    Keyword Arguments:
        apikey: The apikey for authentication [required].
        url: The endpoint for JWT token requests [required].
        scope_collection_type: The scope collection type of item(s) [required].
            Valid values are: "accounts", "subscriptions", "services".
        scope_id: The scope identifier of item(s) [required].
        include_builtin_actions: A flag to include builtin actions in the "actions" claim in the
            MCSP access token (default: False).
        include_custom_actions: A flag to include custom actions in the "actions" claim in the
            MCSP access token (default: False).
        include_roles: A flag to include the "roles" claim in the MCSP access token (default: True).
        prefix_roles: A flag to add a prefix with the scope level where the role is defined
            in the "roles" claim (default: False).
        caller_ext_claim: A dictionary (map) containing keys and values to be injected into
            the access token as the "callerExt" claim (default: None).
            The keys used in this map must be enabled in the apikey by setting the
            "callerExtClaimNames" property when the apikey is created.
        disable_ssl_verification: Disable ssl verification. Defaults to False.
        headers: Headers to be sent with every service token request. Defaults to None.
        proxies: Proxies to use for making request. Defaults to None.
        proxies.http (optional): The proxy endpoint to use for HTTP requests.
        proxies.https (optional): The proxy endpoint to use for HTTPS requests.
    """

    # pylint: disable=too-many-instance-attributes

    # The name of the response body property that contains the access token.
    TOKEN_NAME = 'token'

    # The path associated with the token-exchange operation to be invoked.
    OPERATION_PATH = '/api/2.0/{scopeCollectionType}/{scopeId}/apikeys/token'

    # These path parameter names must be kept in sync with the operation path above.
    _path_param_names = ['scopeCollectionType', 'scopeId']

    def __init__(
        self,
        *,
        apikey: str,
        url: str,
        scope_collection_type: str,
        scope_id: str,
        include_builtin_actions: bool = False,
        include_custom_actions: bool = False,
        include_roles: bool = True,
        prefix_roles: bool = False,
        caller_ext_claim: Optional[Dict[str, str]] = None,
        disable_ssl_verification: bool = False,
        headers: Optional[Dict[str, str]] = None,
        proxies: Optional[Dict[str, str]] = None,
    ) -> None:
        self.apikey = apikey
        self.scope_collection_type = scope_collection_type
        self.scope_id = scope_id
        self.include_builtin_actions = include_builtin_actions
        self.include_custom_actions = include_custom_actions
        self.include_roles = include_roles
        self.prefix_roles = prefix_roles
        self.caller_ext_claim = caller_ext_claim

        self.set_headers(headers)
        self.set_proxies(proxies)

        super().__init__(url, disable_ssl_verification=disable_ssl_verification, token_name=self.TOKEN_NAME)
        self._set_user_agent(_build_user_agent('mcspv2-authenticator'))

    def set_headers(self, headers: Optional[Dict[str, str]] = None) -> None:
        """Headers to be sent with every MCSP token request.

        Args:
            headers: The headers to be sent with every MCSP token request (default: None).
        """
        if isinstance(headers, (dict, type(None))):
            self.headers = headers
        else:
            raise TypeError('"headers" must be a dictionary or None')

    def set_proxies(self, proxies: Optional[Dict[str, str]] = None) -> None:
        """Sets the proxies the token manager will use to communicate with MCSP on behalf of the host.

        Args:
            proxies: Proxies to use for making request (default: None).
            proxies.http (optional): The proxy endpoint to use for HTTP requests.
            proxies.https (optional): The proxy endpoint to use for HTTPS requests.
        """
        if isinstance(proxies, (dict, type(None))):
            self.proxies = proxies
        else:
            raise TypeError('"proxies" must be a dictionary or None')

    def request_token(self) -> dict:
        """Invokes the "POST /api/2.0/{scopeCollectionType}/{scopeId}/apikeys/token" operation
        to obtain an access token."""

        # These headers take priority over user-supplied headers.
        required_headers = {
            'User-Agent': self.user_agent,
            'Content-Type': 'application/json',
            'Accept': 'application/json',
        }

        # Set up the request headers.
        request_headers = {}
        if self.headers is not None and isinstance(self.headers, dict):
            request_headers.update(self.headers)
        request_headers.update(required_headers)

        # Compute the request URL.
        path_param_values = self._encode_path_vars(self.scope_collection_type, self.scope_id)
        path_params = dict(zip(self._path_param_names, path_param_values))
        request_url = self.url + self.OPERATION_PATH.format(**path_params)

        # Create the request body (apikey, callerExtClaim properties).
        request_body = {}
        request_body['apikey'] = self.apikey
        if self.caller_ext_claim is not None and isinstance(self.caller_ext_claim, dict):
            request_body['callerExtClaim'] = self.caller_ext_claim

        # Set up the query params.
        query_params = {
            'includeBuiltinActions': self.bool_to_string(self.include_builtin_actions),
            'includeCustomActions': self.bool_to_string(self.include_custom_actions),
            'includeRoles': self.bool_to_string(self.include_roles),
            'prefixRolesWithDefinitionScope': self.bool_to_string(self.prefix_roles),
        }

        logger.debug('Invoking MCSP v2 token service operation: %s', request_url)

        response = self._request(
            method='POST',
            headers=request_headers,
            url=request_url,
            params=query_params,
            data=json.dumps(request_body),
            proxies=self.proxies,
        )
        logger.debug('Returned from MCSP v2 token service operation')
        return response

    def _encode_path_vars(self, *args: str) -> List[str]:
        """Encode path variables to be substituted into a URL path.

        Arguments:
            args: A list of strings to be URL path encoded

        Returns:
            A list of encoded strings that are safe to substitute into a URL path.
        """
        return (requests.utils.quote(x, safe='') for x in args)

    def bool_to_string(self, value: bool) -> str:
        """Convert a boolean value to string."""
        return 'true' if value else 'false'



# coding: utf-8

# Copyright 2024 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional

from ibm_cloud_sdk_core.authenticators.iam_authenticator import IAMAuthenticator
from ibm_cloud_sdk_core.token_managers.iam_assume_token_manager import IAMAssumeTokenManager

from .authenticator import Authenticator
from .iam_request_based_authenticator import IAMRequestBasedAuthenticator


class IAMAssumeAuthenticator(IAMRequestBasedAuthenticator):
    """IAMAssumeAuthenticator obtains an IAM access token using the IAM "get-token" operation's
    "assume" grant type. The authenticator obtains an initial IAM access token from a
    user-supplied apikey, then exchanges this initial IAM access token for another IAM access token
    that has "assumed the identity" of the specified trusted profile.

    The bearer token will be sent as an Authorization header in the form:

        Authorization: Bearer <bearer-token>

    Args:
        apikey: The IAM api key.

    Keyword Args:
        iam_profile_id: the ID of the trusted profile
        iam_profile_crn: the CRN of the trusted profile
        iam_profile_name: the name of the trusted profile (must be used together with `iam_account_id`)
        iam_account_id: the ID of the trusted profile (must be used together with `iam_profile_name`)
        url: The URL representing the IAM token service endpoint. If not specified, a suitable default value is used.
        client_id: The client_id and client_secret fields are used to form
            a "basic" authorization header for IAM token requests. Defaults to None.
        client_secret: The client_id and client_secret fields are used to form
            a "basic" authorization header for IAM token requests. Defaults to None.
        disable_ssl_verification: A flag that indicates whether verification of
            the server's SSL certificate should be disabled or not. Defaults to False.
        headers: Default headers to be sent with every IAM token request. Defaults to None.
        proxies: Dictionary for mapping request protocol to proxy URL. Defaults to None.
        proxies.http (optional): The proxy endpoint to use for HTTP requests.
        proxies.https (optional): The proxy endpoint to use for HTTPS requests.
        scope: The "scope" to use when fetching the bearer token from the IAM token server.
            This can be used to obtain an access token with a specific scope.

    Attributes:
        token_manager (IAMTokenManager): Retrieves and manages IAM tokens from the endpoint specified by the url.

    Raises:
        TypeError: The `disable_ssl_verification` is not a bool.
        ValueError: The `apikey`, `client_id`, and/or `client_secret` are not valid for IAM token requests or the
            following keyword arguments are incorrectly specified:
            `iam_profile_id`, `iam_profile_crn`, `iam_profile_name`, `iam_account_id`.
    """

    def __init__(
        self,
        apikey: str,
        *,
        iam_profile_id: Optional[str] = None,
        iam_profile_crn: Optional[str] = None,
        iam_profile_name: Optional[str] = None,
        iam_account_id: Optional[str] = None,
        url: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        disable_ssl_verification: bool = False,
        headers: Optional[Dict[str, str]] = None,
        proxies: Optional[Dict[str, str]] = None,
        scope: Optional[str] = None,
    ) -> None:
        # Check the type of `disable_ssl_verification`. Must be a bool.
        if not isinstance(disable_ssl_verification, bool):
            raise TypeError('disable_ssl_verification must be a bool')

        self.token_manager = IAMAssumeTokenManager(
            apikey,
            iam_profile_id=iam_profile_id,
            iam_profile_crn=iam_profile_crn,
            iam_profile_name=iam_profile_name,
            iam_account_id=iam_account_id,
            url=url,
            client_id=client_id,
            client_secret=client_secret,
            disable_ssl_verification=disable_ssl_verification,
            headers=headers,
            proxies=proxies,
            scope=scope,
        )

        self.validate()

    # Disable all setter methods, inherited from the parent class.
    def __getattribute__(self, name: str) -> Any:
        if name.startswith("set_"):
            raise AttributeError(f"'{self.__class__.__name__}' has no attribute '{name}'")

        return super().__getattribute__(name)

    def authentication_type(self) -> str:
        """Returns this authenticator's type ('iamAssume')."""
        return Authenticator.AUTHTYPE_IAM_ASSUME

    def validate(self) -> None:
        """Validates the provided IAM related arguments.

        Ensure the following:
        - `apikey` of the IAMTokenManager is not `None`, and has no bad characters
        - both `client_id` and `client_secret` are set if either of them are defined
        - the correct number and type of IAM profile and IAM account options are specified

        Raises:
            ValueError: The apikey, client_id, and/or client_secret are not valid for IAM token requests.
        """
        # Create a temporary IAM authenticator that we can use to validate our delegate.
        tmp_authenticator = IAMAuthenticator("")
        tmp_authenticator.token_manager = self.token_manager.iam_delegate
        tmp_authenticator.validate()
        del tmp_authenticator

        # Only one of the following arguments must be specified.
        mutually_exclusive_attributes = [
            self.token_manager.iam_profile_id,
            self.token_manager.iam_profile_crn,
            self.token_manager.iam_profile_name,
        ]
        if list(map(bool, mutually_exclusive_attributes)).count(True) != 1:
            raise ValueError(
                'Exactly one of `iam_profile_id`, `iam_profile_crn`, or `iam_profile_name` must be specified.'
            )

        # `iam_account_id` must be specified iff `iam_profile_name` is used.
        if self.token_manager.iam_profile_name and not self.token_manager.iam_account_id:
            raise ValueError('`iam_profile_name` and `iam_account_id` must be provided together, or not at all.')



# coding: utf-8

# Copyright 2019, 2024 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from requests import Request

from ibm_cloud_sdk_core.logger import get_logger
from .authenticator import Authenticator

logger = get_logger()


class IAMRequestBasedAuthenticator(Authenticator):
    """The IAMRequestBasedAuthenticator class contains code that is common to all authenticators
    that need to interact with the IAM tokens service to obtain an access token.

    The bearer token will be sent as an Authorization header in the form:

        Authorization: Bearer <bearer-token>

    Attributes:
        token_manager (TokenManager): Retrieves and manages IAM tokens from the endpoint specified by the url.
    """

    def validate(self) -> None:
        """Validates the client_id, and client_secret for IAM token requests.

        Ensure both the client_id and client_secret are set if either of them are defined.

        Raises:
            ValueError: The client_id, and/or client_secret are not valid for IAM token requests.
        """
        if (self.token_manager.client_id and not self.token_manager.client_secret) or (
            not self.token_manager.client_id and self.token_manager.client_secret
        ):
            raise ValueError('Both client_id and client_secret should be initialized.')

    def authenticate(self, req: Request) -> None:
        """Adds IAM authentication information to the request.

        The IAM bearer token will be added to the request's headers in the form:

            Authorization: Bearer <bearer-token>

        Args:
            req: The request to add IAM authentication information to. Must contain a key to a dictionary
            called headers.
        """
        headers = req.get('headers')
        bearer_token = self.token_manager.get_token()
        headers['Authorization'] = 'Bearer {0}'.format(bearer_token)
        logger.debug('Authenticated outbound request (type=%s)', self.authentication_type())

    def set_client_id_and_secret(self, client_id: str, client_secret: str) -> None:
        """Set the client_id and client_secret pair the token manager will use for IAM token requests.

        Args:
            client_id: The client id to be used in basic auth.
            client_secret: The client secret to be used in basic auth.

        Raises:
            ValueError: The apikey, client_id, and/or client_secret are not valid for IAM token requests.
        """
        self.token_manager.set_client_id_and_secret(client_id, client_secret)
        self.validate()

    def set_disable_ssl_verification(self, status: bool = False) -> None:
        """Set the flag that indicates whether verification of the server's SSL certificate should be
        disabled or not. Defaults to False.

        Args:
            status: Headers to be sent with every IAM token request. Defaults to None

        Raises:
            TypeError: The `status` is not a bool.
        """
        self.token_manager.set_disable_ssl_verification(status)

    def set_headers(self, headers: Dict[str, str]) -> None:
        """Headers to be sent with every IAM token request.

        Args:
            headers: Headers to be sent with every IAM token request.
        """
        self.token_manager.set_headers(headers)

    def set_proxies(self, proxies: Dict[str, str]) -> None:
        """Sets the proxies the token manager will use to communicate with IAM on behalf of the host.

        Args:
            proxies: Dictionary for mapping request protocol to proxy URL.
            proxies.http (optional): The proxy endpoint to use for HTTP requests.
            proxies.https (optional): The proxy endpoint to use for HTTPS requests.
        """
        self.token_manager.set_proxies(proxies)

    def set_scope(self, value: str) -> None:
        """Sets the "scope" parameter to use when fetching the bearer token from the IAM token server.
        This can be used to obtain an access token with a specific scope.

        Args:
            value: A space seperated string that makes up the scope parameter.
        """
        self.token_manager.set_scope(value)



# pylint: disable=missing-docstring
# coding: utf-8

# Copyright 2019, 2024 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging
import os
from typing import Optional

import pytest

from ibm_cloud_sdk_core import string_to_datetime, datetime_to_string
from ibm_cloud_sdk_core import string_to_datetime_list, datetime_to_string_list
from ibm_cloud_sdk_core import string_to_date, date_to_string
from ibm_cloud_sdk_core import convert_model, convert_list
from ibm_cloud_sdk_core import get_query_param
from ibm_cloud_sdk_core import read_external_sources
from ibm_cloud_sdk_core.utils import GzipStream, strip_extra_slashes, is_json_mimetype
from .utils.logger_utils import setup_test_logger

setup_test_logger(logging.ERROR)


def datetime_test(datestr: str, expected: str):
    dt_value = string_to_datetime(datestr)
    assert dt_value is not None
    actual = datetime_to_string(dt_value)
    assert actual == expected


def test_datetime():
    # RFC 3339 with various flavors of tz-offset
    datetime_test('2016-06-20T04:25:16.218Z', '2016-06-20T04:25:16.218000Z')
    datetime_test('2016-06-20T04:25:16.218+0000', '2016-06-20T04:25:16.218000Z')
    datetime_test('2016-06-20T04:25:16.218+00', '2016-06-20T04:25:16.218000Z')
    datetime_test('2016-06-20T04:25:16.218-0000', '2016-06-20T04:25:16.218000Z')
    datetime_test('2016-06-20T04:25:16.218-00', '2016-06-20T04:25:16.218000Z')
    datetime_test('2016-06-20T00:25:16.218-0400', '2016-06-20T04:25:16.218000Z')
    datetime_test('2016-06-20T00:25:16.218-04', '2016-06-20T04:25:16.218000Z')
    datetime_test('2016-06-20T07:25:16.218+0300', '2016-06-20T04:25:16.218000Z')
    datetime_test('2016-06-20T07:25:16.218+03', '2016-06-20T04:25:16.218000Z')
    datetime_test('2016-06-20T04:25:16Z', '2016-06-20T04:25:16Z')
    datetime_test('2016-06-20T04:25:16+0000', '2016-06-20T04:25:16Z')
    datetime_test('2016-06-20T04:25:16-0000', '2016-06-20T04:25:16Z')
    datetime_test('2016-06-20T01:25:16-0300', '2016-06-20T04:25:16Z')
    datetime_test('2016-06-20T01:25:16-03:00', '2016-06-20T04:25:16Z')
    datetime_test('2016-06-20T08:55:16+04:30', '2016-06-20T04:25:16Z')
    datetime_test('2016-06-20T16:25:16+12:00', '2016-06-20T04:25:16Z')

    # RFC 3339 with nanoseconds for the Catalog-Managements of the world.
    datetime_test('2020-03-12T10:52:12.866305005-04:00', '2020-03-12T14:52:12.866305Z')
    datetime_test('2020-03-12T10:52:12.866305005Z', '2020-03-12T10:52:12.866305Z')
    datetime_test('2020-03-12T10:52:12.866305005+02:30', '2020-03-12T08:22:12.866305Z')
    datetime_test('2020-03-12T10:52:12.866305Z', '2020-03-12T10:52:12.866305Z')

    # UTC datetime with no TZ.
    datetime_test('2016-06-20T04:25:16.218', '2016-06-20T04:25:16.218000Z')
    datetime_test('2016-06-20T04:25:16', '2016-06-20T04:25:16Z')

    # Dialog datetime.
    datetime_test('2016-06-20 04:25:16', '2016-06-20T04:25:16Z')

    # IAM Identity Service.
    datetime_test('2020-11-10T12:28+0000', '2020-11-10T12:28:00Z')
    datetime_test('2020-11-10T07:28-0500', '2020-11-10T12:28:00Z')
    datetime_test('2020-11-10T12:28Z', '2020-11-10T12:28:00Z')


def test_string_to_datetime():
    # If the specified string does not include a timezone, it is assumed to be UTC
    date = string_to_datetime('2017-03-06 16:00:04.159338')
    assert date.day == 6
    assert date.hour == 16
    assert date.tzinfo.utcoffset(None) == datetime.timezone.utc.utcoffset(None)
    # Test date string with TZ specified as '+xxxx'
    date = string_to_datetime('2017-03-06 16:00:04.159338+0600')
    assert date.day == 6
    assert date.hour == 16
    assert date.tzinfo.utcoffset(None).total_seconds() == 6 * 60 * 60
    # Test date string with TZ specified as 'Z'
    date = string_to_datetime('2017-03-06 16:00:04.159338Z')
    assert date.day == 6
    assert date.hour == 16
    assert date.tzinfo.utcoffset(None) == datetime.timezone.utc.utcoffset(None)


def test_datetime_to_string():
    # If specified date is None, return None
    assert datetime_to_string(None) is None
    # If the specified date is "naive", it is interpreted as a UTC date
    date = datetime.datetime(2017, 3, 6, 16, 0, 4, 159338)
    res = datetime_to_string(date)
    assert res == '2017-03-06T16:00:04.159338Z'
    # Test date with UTC timezone
    date = datetime.datetime(2017, 3, 6, 16, 0, 4, 159338, datetime.timezone.utc)
    res = datetime_to_string(date)
    assert res == '2017-03-06T16:00:04.159338Z'
    # Test date with non-UTC timezone
    tzn = datetime.timezone(datetime.timedelta(hours=-6))
    date = datetime.datetime(2017, 3, 6, 10, 0, 4, 159338, tzn)
    res = datetime_to_string(date)
    assert res == '2017-03-06T16:00:04.159338Z'


def test_string_to_datetime_list():
    # Assert ValueError is raised for invalid argument type
    with pytest.raises(ValueError):
        string_to_datetime_list(None)
    # If the specified string does not include a timezone, it is assumed to be UTC
    date_list = string_to_datetime_list(['2017-03-06 16:00:04.159338'])
    assert date_list[0].day == 6
    assert date_list[0].hour == 16
    assert date_list[0].tzinfo.utcoffset(None) == datetime.timezone.utc.utcoffset(None)
    # Test date string with TZ specified as '+xxxx'
    date_list = string_to_datetime_list(['2017-03-06 16:00:04.159338+0600'])
    assert date_list[0].day == 6
    assert date_list[0].hour == 16
    assert date_list[0].tzinfo.utcoffset(None).total_seconds() == 6 * 60 * 60
    # Test date string with TZ specified as 'Z'
    date_list = string_to_datetime_list(['2017-03-06 16:00:04.159338Z'])
    assert date_list[0].day == 6
    assert date_list[0].hour == 16
    assert date_list[0].tzinfo.utcoffset(None) == datetime.timezone.utc.utcoffset(None)
    # Test multiple datetimes in a list
    date_list = string_to_datetime_list(['2017-03-06 16:00:04.159338', '2017-03-07 17:00:04.159338'])
    assert date_list[0].day == 6
    assert date_list[0].hour == 16
    assert date_list[0].tzinfo.utcoffset(None) == datetime.timezone.utc.utcoffset(None)
    assert date_list[1].day == 7
    assert date_list[1].hour == 17
    assert date_list[1].tzinfo.utcoffset(None) == datetime.timezone.utc.utcoffset(None)


def test_datetime_to_string_list():
    # Assert ValueError is raised for invalid argument type
    with pytest.raises(ValueError):
        datetime_to_string_list(None)
    # If specified datetime list item is None, return list of None
    assert datetime_to_string_list([None]) == [None]
    # If specified datetime list is empty, return empty list
    assert not datetime_to_string_list([])
    # If the specified date list item is "naive", it is interpreted as a UTC date
    date_list = [datetime.datetime(2017, 3, 6, 16, 0, 4, 159338)]
    res = datetime_to_string_list(date_list)
    assert res == ['2017-03-06T16:00:04.159338Z']
    # Test date list item with UTC timezone
    date_list = [datetime.datetime(2017, 3, 6, 16, 0, 4, 159338, datetime.timezone.utc)]
    res = datetime_to_string_list(date_list)
    assert res == ['2017-03-06T16:00:04.159338Z']
    # Test date list item with non-UTC timezone
    tzn = datetime.timezone(datetime.timedelta(hours=-6))
    date_list = [datetime.datetime(2017, 3, 6, 10, 0, 4, 159338, tzn)]
    res = datetime_to_string_list(date_list)
    assert res == ['2017-03-06T16:00:04.159338Z']
    # Test specified date list with multiple items
    date_list = [
        datetime.datetime(2017, 3, 6, 16, 0, 4, 159338),
        datetime.datetime(2017, 3, 6, 16, 0, 4, 159338, datetime.timezone.utc),
    ]
    res = datetime_to_string_list(date_list)
    assert res == ['2017-03-06T16:00:04.159338Z', '2017-03-06T16:00:04.159338Z']


def test_date_conversion():
    date = string_to_date('2017-03-06')
    assert date.day == 6
    res = date_to_string(date)
    assert res == '2017-03-06'
    assert date_to_string(None) is None


def test_get_query_param():
    # Relative URL
    next_url = '/api/v1/offerings?start=foo&limit=10'
    page_token = get_query_param(next_url, 'start')
    assert page_token == 'foo'
    # Absolute URL
    next_url = 'https://acme.com/api/v1/offerings?start=bar&limit=10'
    page_token = get_query_param(next_url, 'start')
    assert page_token == 'bar'
    # Missing param
    next_url = 'https://acme.com/api/v1/offerings?start=bar&limit=10'
    page_token = get_query_param(next_url, 'token')
    assert page_token is None
    # No URL
    page_token = get_query_param(None, 'start')
    assert page_token is None
    # Empty URL
    page_token = get_query_param('', 'start')
    assert page_token is None
    # No query string
    next_url = '/api/v1/offerings'
    page_token = get_query_param(next_url, 'start')
    assert page_token is None
    # Bad query string
    next_url = '/api/v1/offerings?start%XXfoo'
    with pytest.raises(ValueError):
        page_token = get_query_param(next_url, 'start')
    # Duplicate param
    next_url = '/api/v1/offerings?start=foo&start=bar&limit=10'
    page_token = get_query_param(next_url, 'start')
    assert page_token == 'foo'
    # Bad URL - since the behavior for this case varies based on the version of Python
    # we allow _either_ a ValueError or that the illegal chars are just ignored
    next_url = 'https://foo.bar\u2100/api/v1/offerings?start=foo'
    try:
        page_token = get_query_param(next_url, 'start')
        assert page_token == 'foo'
    except ValueError:
        # This is okay.
        pass


def test_convert_model():
    class MockModel:
        def __init__(self, xyz: Optional[str] = None) -> None:
            self.xyz = xyz

        def to_dict(self) -> dict:
            _dict = {}
            if hasattr(self, 'xyz') and self.xyz is not None:
                _dict['xyz'] = self.xyz
            return _dict

        @classmethod
        def from_dict(cls, _dict):
            pass

    mock1 = MockModel('foo')
    mock1_dict = convert_model(mock1)
    assert mock1_dict == {'xyz': 'foo'}

    mock2 = {'foo': 'bar', 'baz': 'qux'}
    mock2_dict = convert_model(mock2)
    assert mock2_dict == mock2

    mock3 = 'this is not a model'
    mock3_dict = convert_model(mock3)
    assert mock3_dict == mock3


def test_convert_list():
    temp = ['default', '123']
    res_str = convert_list(temp)
    assert res_str == 'default,123'

    mock2 = 'default,123'
    mock2_str = convert_list(mock2)
    assert mock2_str == mock2

    mock3 = {'not': 'a list'}
    mock3_str = convert_list(mock3)
    assert mock3_str == mock3

    mock4 = ['not', 0, 'list of str']
    mock4_str = convert_list(mock4)
    assert mock4_str == mock4


def test_read_external_sources_1():
    # Set IBM_CREDENTIALS_FILE to a non-existent file (should be silently ignored).
    bad_file_path = os.path.join(os.path.dirname(__file__), 'NOT_A_FILE')
    os.environ['IBM_CREDENTIALS_FILE'] = bad_file_path

    # This env var should take precendence since the config file wasn't found.
    os.environ['SERVICE_1_URL'] = 'https://good-url.com'

    config = read_external_sources('service_1')
    assert config.get('URL') == 'https://good-url.com'


def test_read_external_sources_2():
    # The config file should take precedence over the env variable.
    config_file = os.path.join(os.path.dirname(__file__), '../resources/ibm-credentials.env')
    os.environ['IBM_CREDENTIALS_FILE'] = config_file

    # This should be ignored since IBM_CREDENTIALS_FILE points to a valid file.
    os.environ['SERVICE_1_URL'] = 'wrong-url'

    config = read_external_sources('service_1')
    assert config.get('URL') == 'service1.com/api'


def test_strip_extra_slashes():
    assert strip_extra_slashes('') == ''
    assert strip_extra_slashes('//') == '/'
    assert strip_extra_slashes('/////') == '/'
    assert strip_extra_slashes('https://host') == 'https://host'
    assert strip_extra_slashes('https://host/') == 'https://host/'
    assert strip_extra_slashes('https://host//') == 'https://host/'
    assert strip_extra_slashes('https://host/path') == 'https://host/path'
    assert strip_extra_slashes('https://host/path/') == 'https://host/path/'
    assert strip_extra_slashes('https://host/path//') == 'https://host/path/'
    assert strip_extra_slashes('https://host//path//') == 'https://host//path/'
    assert strip_extra_slashes('https://host//path//////////') == 'https://host//path/'


def test_is_json_mimetype():
    assert is_json_mimetype(None) is False
    assert is_json_mimetype('') is False
    assert is_json_mimetype('application/octet-stream') is False
    assert is_json_mimetype('ApPlIcAtION/JsoN') is False
    assert is_json_mimetype('applicaiton/json; charset=utf8') is False
    assert is_json_mimetype('fooapplication/json; charset=utf8; foo=bar') is False

    assert is_json_mimetype('application/json') is True
    assert is_json_mimetype('application/json; charset=utf8') is True


def test_gzip_stream_open_file():
    cr_token_file = os.path.join(os.path.dirname(__file__), '../resources/cr-token.txt')
    with open(cr_token_file, 'r', encoding='utf-8') as f:
        stream = GzipStream(source=f)
        assert stream is not None


def test_gzip_stream_open_string():
    stream = GzipStream(source='foobar')
    assert stream is not None


def test_gzip_stream_open_bytes():
    stream = GzipStream(source=b'foobar')
    assert stream is not None



# coding: utf-8

# Copyright 2019 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from http import HTTPStatus
from typing import Optional

from requests import Response


class ApiException(Exception):
    """Custom exception class for errors returned from operations.

    Args:
        code: HTTP status code of the error response.
        message: The error response body. Defaults to None.
        http_response: The HTTP response of the failed request. Defaults to None.

    Attributes:
        code (int): HTTP status code of the error response.
        message (str): The error response body.
        http_response (requests.Response): The HTTP response of the failed request.
        global_transaction_id (str, optional): Globally unique id the service endpoint has given a transaction.
    """

    def __init__(self, code: int, *, message: Optional[str] = None, http_response: Optional[Response] = None) -> None:
        # Call the base class constructor with the parameters it needs
        super().__init__(message)
        self.message = message
        self.status_code = code
        self.http_response = http_response
        self.global_transaction_id = None
        if http_response is not None:
            self.global_transaction_id = http_response.headers.get('X-Global-Transaction-ID')
            self.message = self.message if self.message else self._get_error_message(http_response)

    # pylint: disable=fixme
    # TODO: delete this by the end of 2024.
    @property
    def code(self):
        """The old `code` property with a deprecation warning."""

        warnings.warn(
            'Using the `code` attribute on the `ApiException` is deprecated and '
            'will be removed in the future. Use `status_code` instead.',
            DeprecationWarning,
        )
        return self.status_code

    def __str__(self) -> str:
        msg = 'Error: ' + str(self.message) + ', Status code: ' + str(self.status_code)
        if self.global_transaction_id is not None:
            msg += ' , X-global-transaction-id: ' + str(self.global_transaction_id)
        return msg

    @staticmethod
    def _get_error_message(response: Response) -> str:
        error_message = 'Unknown error'
        try:
            error_json = response.json(strict=False)
            if 'errors' in error_json:
                if isinstance(error_json['errors'], list):
                    err = error_json['errors'][0]
                    error_message = err.get('message')
            elif 'error' in error_json:
                error_message = error_json['error']
            elif 'message' in error_json:
                error_message = error_json['message']
            elif 'errorMessage' in error_json:
                error_message = error_json['errorMessage']
            elif response.status_code == 401:
                error_message = 'Unauthorized: Access is denied due to invalid credentials'
            else:
                error_message = HTTPStatus(response.status_code).phrase
            return error_message
        except:
            return response.text or error_message



# coding: utf-8

# Copyright 2021, 2024 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Optional

from ibm_cloud_sdk_core.logger import get_logger
from ..private_helpers import _build_user_agent
from .jwt_token_manager import JWTTokenManager

logger = get_logger()


class VPCInstanceTokenManager(JWTTokenManager):
    """The VPCInstanceTokenManager retrieves an "instance identity token" and exchanges that
       for an IAM access token using the VPC Instance Metadata Service API which is available
       on the local compute resource (VM).
       The instance identity token is similar to an IAM apikey, except that it is managed
       automatically by the compute resource provider (VPC).
       The resulting IAM access token is then added to outbound requests in an Authorization header of the form:

           Authorization: Bearer <access-token>

    Keyword Arguments:
        iam_profile_crn (str, optional):
            The CRN of the linked trusted IAM profile to be used as the identity of the compute resource.
            At most one of iam_profile_crn or iam_profile_id may be specified. If neither one is specified,
            then the default IAM profile defined for the compute resource will be used. Defaults to None.
        iam_profile_id (str, optional):
            The ID of the linked trusted IAM profile to be used when obtaining the IAM access token.
            At most one of iam_profile_crn or iam_profile_id may be specified. If neither one is specified,
            then the default IAM profile defined for the compute resource will be used. Defaults to None.
        url (str, optional):
            The VPC Instance Metadata Service's base endpoint URL. Defaults to 'http://169.254.169.254'.

    Attributes:
        iam_profile_crn (str, optional): The CRN of the linked trusted IAM profile.
        iam_profile_id (str, optional): The ID of the linked trusted IAM profile.
        url (str, optional): The VPC Instance Metadata Service's base endpoint URL.
    """

    METADATA_SERVICE_VERSION = '2022-03-01'
    DEFAULT_IMS_ENDPOINT = 'http://169.254.169.254'
    TOKEN_NAME = 'access_token'
    IAM_EXPIRATION_WINDOW = 10

    def __init__(
        self, iam_profile_crn: Optional[str] = None, iam_profile_id: Optional[str] = None, url: Optional[str] = None
    ) -> None:
        if not url:
            url = self.DEFAULT_IMS_ENDPOINT

        super().__init__(url, token_name=self.TOKEN_NAME)
        self._set_user_agent(_build_user_agent('vpc-instance-authenticator'))

        self.iam_profile_crn = iam_profile_crn
        self.iam_profile_id = iam_profile_id

    def request_token(self) -> dict:
        """RequestToken will use the VPC Instance Metadata Service to
           (1) retrieve a fresh instance identity token and then
           (2) exchange that for an IAM access token.

        Returns:
            A dictionary containing the bearer token to be subsequently used service requests.
        """

        # Retrieve the Instance Identity Token first.
        instance_identity_token = self.retrieve_instance_identity_token()

        url = self.url + '/instance_identity/v1/iam_token'

        request_payload = None
        if self.iam_profile_crn:
            request_payload = {'trusted_profile': {'crn': self.iam_profile_crn}}
        if self.iam_profile_id:
            request_payload = {'trusted_profile': {'id': self.iam_profile_id}}

        headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Authorization': 'Bearer ' + instance_identity_token,
            'User-Agent': self._get_user_agent(),
        }

        logger.debug('Invoking VPC \'create_iam_token\' operation: %s', url)
        response = self._request(
            method='POST',
            url=url,
            headers=headers,
            params={'version': self.METADATA_SERVICE_VERSION},
            data=json.dumps(request_payload) if request_payload else None,
        )
        logger.debug('Returned from VPC \'create_iam_token\' operation.')

        return response

    def set_iam_profile_crn(self, iam_profile_crn: str) -> None:
        """Sets the CRN of the IAM profile.

        Args:
            iam_profile_crn (str): the CRN of the linked trusted IAM profile to be used as
                             the identity of the compute resource.
        """
        self.iam_profile_crn = iam_profile_crn

    def set_iam_profile_id(self, iam_profile_id: str) -> None:
        """Sets the ID of the IAM profile.

        Args:
            iam_profile_id (str): id of the linked trusted IAM profile to be used when obtaining
                            the IAM access token
        """
        self.iam_profile_id = iam_profile_id

    def retrieve_instance_identity_token(self) -> str:
        """Retrieves the local compute resource's instance identity token using
           the "create_access_token" operation of the local VPC Instance Metadata Service API.

        Returns:
            The retrieved instance identity token string.
        """

        url = self.url + '/instance_identity/v1/token'

        headers = {
            'Content-type': 'application/json',
            'Accept': 'application/json',
            'Metadata-Flavor': 'ibm',
            'User-Agent': self._get_user_agent(),
        }

        request_body = {'expires_in': 300}

        logger.debug('Invoking VPC \'create_access_token\' operation: %s', url)
        response = self._request(
            method='PUT',
            url=url,
            headers=headers,
            params={'version': self.METADATA_SERVICE_VERSION},
            data=json.dumps(request_body),
        )
        logger.debug('Returned from VPC \'create_access_token\' operation.')

        return response['access_token']

    def _is_token_expired(self) -> bool:
        """
        Returns true iff the current cached token is expired.
        We'll consider an access token as expired when we reach its IAM server-reported expiration time
        minus our expiration window (10 secs).
        We do this to avoid using an access token that might expire in the middle of a long-running transaction
        within an IBM Cloud service.

        Returns
        -------
        bool
            True if token is expired; False otherwise
        """
        current_time = self._get_current_time()
        return current_time >= (self.expire_time - self.IAM_EXPIRATION_WINDOW)



# coding: utf-8

# Copyright 2024 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# from ibm_cloud_sdk_core.authenticators import Authenticator

import platform
from .version import __version__

SDK_NAME = 'ibm-python-sdk-core'


def _get_system_info() -> str:
    return 'os.name={0} os.version={1} python.version={2}'.format(
        platform.system(), platform.release(), platform.python_version()
    )


def _build_user_agent(component: str = None) -> str:
    sub_component = ""
    if component is not None:
        sub_component = '/{0}'.format(component)
    return '{0}{1}-{2} {3}'.format(SDK_NAME, sub_component, __version__, _get_system_info())



# coding: utf-8

# Copyright 2019, 2024 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Dict, Optional

from ibm_cloud_sdk_core.logger import get_logger
from ..private_helpers import _build_user_agent
from .jwt_token_manager import JWTTokenManager

logger = get_logger()


class CP4DTokenManager(JWTTokenManager):
    """Token Manager of CloudPak for data.

    The Token Manager performs basic auth with a username and password
    to acquire JWT tokens.

    Keyword Arguments:
        username: The username for authentication [required].
        password: The password for authentication [required if apikey not specified].
        url: The endpoint for JWT token requests [required].
        apikey: The apikey for authentication [required if password not specified].
        disable_ssl_verification: Disable ssl verification. Defaults to False.
        headers: Headers to be sent with every service token request. Defaults to None.
        proxies: Proxies to use for making request. Defaults to None.
        proxies.http (optional): The proxy endpoint to use for HTTP requests.
        proxies.https (optional): The proxy endpoint to use for HTTPS requests.
        verify (optional): The path to the certificate to use for HTTPS requests.

    Attributes:
        username (str): The username for authentication.
        password (str): The password for authentication.
        url (str): The endpoint for JWT token requests.
        headers (dict): Headers to be sent with every service token request.
        proxies (dict): Proxies to use for making token requests.
        proxies.http (str): The proxy endpoint to use for HTTP requests.
        proxies.https (str): The proxy endpoint to use for HTTPS requests.
        verify (str): The path to the certificate to use for HTTPS requests.
    """

    TOKEN_NAME = 'token'
    VALIDATE_AUTH_PATH = '/v1/authorize'

    def __init__(
        self,
        username: str = None,
        password: str = None,
        url: str = None,
        *,
        apikey: str = None,
        disable_ssl_verification: bool = False,
        headers: Optional[Dict[str, str]] = None,
        proxies: Optional[Dict[str, str]] = None,
        verify: Optional[str] = None,
    ) -> None:
        self.username = username
        self.password = password
        self.verify = verify
        if url and not self.VALIDATE_AUTH_PATH in url:
            url = url + '/v1/authorize'
        self.apikey = apikey
        self.headers = headers
        if self.headers is None:
            self.headers = {}
        self.headers['Content-Type'] = 'application/json'
        self.proxies = proxies
        super().__init__(url, disable_ssl_verification=disable_ssl_verification, token_name=self.TOKEN_NAME)
        self._set_user_agent(_build_user_agent('cp4d-authenticator'))

    def request_token(self) -> dict:
        """Makes a request for a token."""
        required_headers = {
            'User-Agent': self.user_agent,
        }
        request_headers = {}
        if self.headers is not None and isinstance(self.headers, dict):
            request_headers.update(self.headers)
        request_headers.update(required_headers)

        logger.debug('Invoking CP4D token service operation: %s', self.url)
        response = self._request(
            method='POST',
            headers=request_headers,
            url=self.url,
            data=json.dumps({"username": self.username, "password": self.password, "api_key": self.apikey}),
            proxies=self.proxies,
            verify=self.verify,
        )
        logger.debug('Returned from CP4D token service operation')
        return response

    def set_headers(self, headers: Dict[str, str]) -> None:
        """Headers to be sent with every CP4D token request.

        Args:
            headers: The headers to be sent with every CP4D token request.
        """
        if isinstance(headers, dict):
            self.headers = headers
        else:
            raise TypeError('headers must be a dictionary')

    def set_proxies(self, proxies: Dict[str, str]) -> None:
        """Sets the proxies the token manager will use to communicate with CP4D on behalf of the host.

        Args:
            proxies: Proxies to use for making request. Defaults to None.
            proxies.http (optional): The proxy endpoint to use for HTTP requests.
            proxies.https (optional): The proxy endpoint to use for HTTPS requests.
        """
        if isinstance(proxies, dict):
            self.proxies = proxies
        else:
            raise TypeError('proxies must be a dictionary')



# coding: utf-8

# Copyright 2019, 2024 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gzip
import io
import logging
import json as json_import
from http.cookiejar import CookieJar
from http import client
from os.path import basename
from typing import Dict, List, Optional, Tuple, Union
from urllib3.util.retry import Retry

import requests
from requests.structures import CaseInsensitiveDict
from requests.exceptions import JSONDecodeError

from ibm_cloud_sdk_core.authenticators import Authenticator
from .api_exception import ApiException
from .detailed_response import DetailedResponse
from .http_adapter import SSLHTTPAdapter
from .token_managers.token_manager import TokenManager
from .utils import (
    has_bad_first_or_last_char,
    is_json_mimetype,
    remove_null_values,
    cleanup_values,
    read_external_sources,
    strip_extra_slashes,
    GzipStream,
)
from .private_helpers import _build_user_agent
from .logger import (
    get_logger,
    LoggingFilter,
)

logger = get_logger()


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-locals
class BaseService:
    """Common functionality shared by generated service classes.

    The base service authenticates requests via its authenticator, stores cookies, and
    wraps responses from the service endpoint in DetailedResponse or APIException objects.

    Keyword Arguments:
        service_url: Url to the service endpoint. Defaults to None.
        authenticator: Adds authentication data to service requests. Defaults to None.
        disable_ssl_verification: A flag that indicates whether verification of the server's SSL
            certificate should be disabled or not. Defaults to False.
        enable_gzip_compression: A flag that indicates whether to enable gzip compression on request bodies

    Attributes:
        service_url (str): Url to the service endpoint.
        authenticator (Authenticator): Adds authentication data to service requests.
        disable_ssl_verification (bool): A flag that indicates whether verification of
            the server's SSL certificate should be disabled or not.
        default_headers (dict): A dictionary of headers to be sent with every HTTP request to the service endpoint.
        jar (http.cookiejar.CookieJar): Stores cookies received from the service.
        http_config (dict): A dictionary containing values that control the timeout, proxies, and etc of HTTP requests.
        http_client (Session): A configurable session which can use Transport Adapters to configure retries, timeouts,
            proxies, etc. globally for all requests.
        enable_gzip_compression (bool): A flag that indicates whether to enable gzip compression on request bodies
    Raises:
        ValueError: If Authenticator is not provided or invalid type.
    """

    ERROR_MSG_DISABLE_SSL = (
        'The connection failed because the SSL certificate is not valid. To use a self-signed '
        'certificate, disable verification of the server\'s SSL certificate by invoking the '
        'set_disable_ssl_verification(True) on your service instance and/ or use the '
        'disable_ssl_verification option of the authenticator.'
    )

    def __init__(
        self,
        *,
        service_url: str = None,
        authenticator: Optional[Authenticator] = None,
        disable_ssl_verification: bool = False,
        enable_gzip_compression: bool = False,
    ) -> None:
        self.set_service_url(service_url)
        self.http_client = requests.Session()
        self.http_config = {}
        self.jar = CookieJar()
        self.authenticator = authenticator
        self.disable_ssl_verification = disable_ssl_verification
        self.default_headers = None
        self.enable_gzip_compression = enable_gzip_compression
        self._set_user_agent_header(_build_user_agent())
        self.retry_config = None
        self.http_adapter = SSLHTTPAdapter(_disable_ssl_verification=self.disable_ssl_verification)
        if not self.authenticator:
            raise ValueError('authenticator must be provided')
        if not isinstance(self.authenticator, Authenticator):
            raise ValueError('authenticator should be of type Authenticator')

        self.http_client.mount('http://', self.http_adapter)
        self.http_client.mount('https://', self.http_adapter)
        # If debug logging is requested, then trigger HTTP message logging as well.
        if logger.isEnabledFor(logging.DEBUG):
            client.HTTPConnection.debuglevel = 1
            # Replace the `print` function in the HTTPClient module to
            # use the debug logger instead of the bare Python print.
            client.print = lambda *args: logger.debug(LoggingFilter.filter_message(" ".join(args)))

    def enable_retries(self, max_retries: int = 4, retry_interval: float = 30.0) -> None:
        """Enable automatic retries on the underlying http client used by the BaseService instance.

        Args:
          max_retries: the maximum number of retries to attempt for a failed retryable request
          retry_interval: the maximum wait time (in seconds) to use for retry attempts.
            In general, if a response includes the Retry-After header, that will be used for
            the wait time associated with the retry attempt.  If the Retry-After header is not
            present, then the wait time is based on an exponential backoff policy with a maximum
            backoff time of "retry_interval".
        """
        self.retry_config = Retry(
            total=max_retries,
            backoff_factor=1.0,
            backoff_max=retry_interval,
            # List of HTTP status codes to retry on in addition to Timeout/Connection Errors
            status_forcelist=[429, 500, 502, 503, 504],
            # List of HTTP methods to retry on
            # Omitting this will default to all methods except POST
            allowed_methods=['HEAD', 'GET', 'PUT', 'DELETE', 'OPTIONS', 'TRACE', 'POST'],
        )
        self.http_adapter = SSLHTTPAdapter(
            max_retries=self.retry_config, _disable_ssl_verification=self.disable_ssl_verification
        )
        self.http_client.mount('http://', self.http_adapter)
        self.http_client.mount('https://', self.http_adapter)
        logger.debug('Enabled retries; max_retries=%d, max_retry_interval=%f', max_retries, retry_interval)

    def disable_retries(self):
        """Remove retry config from http_adapter"""
        self.retry_config = None
        self.http_adapter = SSLHTTPAdapter(_disable_ssl_verification=self.disable_ssl_verification)
        self.http_client.mount('http://', self.http_adapter)
        self.http_client.mount('https://', self.http_adapter)
        logger.debug('Disabled retries')

    def configure_service(self, service_name: str) -> None:
        """Look for external configuration of a service. Set service properties.

        Try to get config from external sources, with the following priority:
        1. Credentials file(ibm-credentials.env)
        2. Environment variables
        3. VCAP Services(Cloud Foundry)

        Args:
            service_name: The service name

        Raises:
            ValueError: If service_name is not a string.
        """
        if not isinstance(service_name, str):
            raise ValueError('Service_name must be of type string.')

        logger.debug('Configuring BaseService instance with service name: %s', service_name)

        config = read_external_sources(service_name)
        if config.get('URL'):
            self.set_service_url(config.get('URL'))
        if config.get('DISABLE_SSL'):
            self.set_disable_ssl_verification(config.get('DISABLE_SSL').lower() == 'true')
        if config.get('ENABLE_GZIP'):
            self.set_enable_gzip_compression(config.get('ENABLE_GZIP').lower() == 'true')
        if config.get('ENABLE_RETRIES'):
            if config.get('ENABLE_RETRIES').lower() == 'true':
                kwargs = {}
                if config.get('MAX_RETRIES'):
                    kwargs["max_retries"] = int(config.get('MAX_RETRIES'))
                if config.get('RETRY_INTERVAL'):
                    kwargs["retry_interval"] = float(config.get('RETRY_INTERVAL'))
                self.enable_retries(**kwargs)

    def _set_user_agent_header(self, user_agent_string: str) -> None:
        self.user_agent_header = {'User-Agent': user_agent_string}
        logger.debug('Set User-Agent: %s', user_agent_string)

    def set_http_config(self, http_config: dict) -> None:
        """Sets the http config dictionary.

        The dictionary can contain values that control the timeout, proxies, and etc of HTTP requests.

        Arguments:
            http_config: Configuration values to customize HTTP behaviors.

        Raises:
            TypeError: http_config is not a dict.
        """
        if isinstance(http_config, dict):
            self.http_config = http_config
            if (
                self.authenticator
                and hasattr(self.authenticator, 'token_manager')
                and isinstance(self.authenticator.token_manager, TokenManager)
            ):
                self.authenticator.token_manager.http_config = http_config
        else:
            raise TypeError("http_config parameter must be a dictionary")

    def set_disable_ssl_verification(self, status: bool = False) -> None:
        """Set the flag that indicates whether verification of
        the server's SSL certificate should be disabled or not.

        Keyword Arguments:
            status: set to true to disable ssl verification (default: {False})
        """
        if self.disable_ssl_verification == status:
            # Do nothing if the state doesn't change.
            return

        self.disable_ssl_verification = status

        self.http_adapter = SSLHTTPAdapter(
            max_retries=self.retry_config, _disable_ssl_verification=self.disable_ssl_verification
        )
        self.http_client.mount('http://', self.http_adapter)
        self.http_client.mount('https://', self.http_adapter)
        logger.debug('Disabled SSL verification in HTTP client')

    def set_service_url(self, service_url: str) -> None:
        """Set the url the service will make HTTP requests too.

        Arguments:
            service_url: The WHATWG URL standard origin ex. https://example.service.com

        Raises:
            ValueError: Improperly formatted service_url
        """
        if has_bad_first_or_last_char(service_url):
            raise ValueError(
                'The service url shouldn\'t start or end with curly brackets or quotes. '
                'Be sure to remove any {} and \" characters surrounding your service url'
            )
        if service_url is not None:
            service_url = service_url.rstrip('/')
        self.service_url = service_url
        logger.debug('Set service URL: %s', service_url)

    def get_http_client(self) -> requests.sessions.Session:
        """Get the http client session currently used by the service.

        Returns:
            The http client session currently used by the service.
        """
        return self.http_client

    def set_http_client(self, http_client: requests.sessions.Session) -> None:
        """Set current http client session

        Arguments:
            http_client: A new requests session client
        """
        if isinstance(http_client, requests.sessions.Session):
            self.http_client = http_client
        else:
            raise TypeError("http_client parameter must be a requests.sessions.Session")

    def get_authenticator(self) -> Authenticator:
        """Get the authenticator currently used by the service.

        Returns:
            The authenticator currently used by the service.
        """
        return self.authenticator

    def set_default_headers(self, headers: Dict[str, str]) -> None:
        """Set http headers to be sent in every request.

        Arguments:
            headers: A dictionary of headers
        """
        if isinstance(headers, dict):
            self.default_headers = headers
        else:
            raise TypeError("headers parameter must be a dictionary")

    def send(self, request: requests.Request, **kwargs) -> DetailedResponse:
        """Send a request and wrap the response in a DetailedResponse or APIException.

        Args:
            request: The request to send to the service endpoint.

        Raises:
            ApiException: The exception from the API.

        Returns:
            The response from the request.
        """
        # Use a one minute timeout when our caller doesn't give a timeout.
        # http://docs.python-requests.org/en/master/user/quickstart/#timeouts
        kwargs = dict({"timeout": 60}, **kwargs)
        kwargs = dict(kwargs, **self.http_config)

        if self.disable_ssl_verification:
            kwargs['verify'] = False

        # Check to see if the caller specified the 'stream' argument.
        stream_response = kwargs.get('stream') or False

        # Remove the keys we set manually, don't let the user overwrite these.
        reserved_keys = ['method', 'url', 'headers', 'params', 'cookies']
        silent_keys = ['headers']
        for key in reserved_keys:
            if key in kwargs:
                del kwargs[key]
                if key not in silent_keys:
                    logger.warning('"%s" has been removed from the request', key)
        try:
            logger.debug('Sending HTTP request message')

            response = self.http_client.request(**request, cookies=self.jar, **kwargs)

            logger.debug('Received HTTP response message, status code %d', response.status_code)

            # Process a "success" response.
            if 200 <= response.status_code <= 299:
                if response.status_code == 204 or request['method'] == 'HEAD':
                    # There is no body content for a HEAD response or a 204 response.
                    result = None
                elif stream_response:
                    result = response
                elif not response.text:
                    result = None
                elif is_json_mimetype(response.headers.get('Content-Type')):
                    # If this is a JSON response, then try to unmarshal it.
                    try:
                        result = response.json(strict=False)
                    except JSONDecodeError as err:
                        raise ApiException(
                            code=response.status_code,
                            http_response=response,
                            message='Error processing the HTTP response',
                        ) from err
                else:
                    # Non-JSON response, just use response body as-is.
                    result = response

                return DetailedResponse(response=result, headers=response.headers, status_code=response.status_code)

            # Received error status code from server, raise an APIException.
            raise ApiException(response.status_code, http_response=response)
        except requests.exceptions.SSLError:
            logger.exception(self.ERROR_MSG_DISABLE_SSL)
            raise

    def set_enable_gzip_compression(self, should_enable_compression: bool = False) -> None:
        """Set value to enable gzip compression on request bodies"""
        self.enable_gzip_compression = should_enable_compression

    def get_enable_gzip_compression(self) -> bool:
        """Get value for enabling gzip compression on request bodies"""
        return self.enable_gzip_compression

    def prepare_request(
        self,
        method: str,
        url: str,
        *,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
        data: Optional[Union[str, dict]] = None,
        files: Optional[Union[Dict[str, Tuple[str]], List[Tuple[str, Tuple[str, ...]]]]] = None,
        **kwargs,
    ) -> dict:
        """Build a dict that represents an HTTP service request.

        Clean up headers, add default http configuration, convert data
        into json, process files, and merge all into a single request dict.

        Args:
            method: The HTTP method of the request ex. GET, POST, etc.
            url: The origin + pathname according to WHATWG spec.

        Keyword Arguments:
            headers: A dictionary containing the headers to be included in the request.
                    Entries with a value of None will be ignored (excluded).
            params: A dictionary containing the query parameters to be included in the request.
                    Entries with a value of None will be ignored (excluded).
            data: The request body. Converted to json if a dict.
            files: 'files' can be a dictionary (i.e { '<part-name>': (<tuple>)}),
                or a list of tuples [ (<part-name>, (<tuple>))... ]

        Returns:
            Prepared request dictionary.
        """
        # pylint: disable=unused-argument; necessary for kwargs
        request = {'method': method}

        # validate the service url is set
        if not self.service_url:
            raise ValueError('The service_url is required')

        # Combine the service_url and operation path to form the request url.
        # Note: we have already stripped any trailing slashes from the service_url
        # and we know that the operation path ('url') will start with a slash.
        request['url'] = strip_extra_slashes(self.service_url + url)

        headers = remove_null_values(headers) if headers else {}
        headers = cleanup_values(headers)
        headers = CaseInsensitiveDict(headers)
        if self.default_headers is not None:
            headers.update(self.default_headers)
        if 'user-agent' not in headers:
            headers.update(self.user_agent_header)
        request['headers'] = headers

        params = remove_null_values(params)
        params = cleanup_values(params)
        request['params'] = params

        if isinstance(data, str):
            data = data.encode('utf-8')
        elif isinstance(data, dict) and data:
            data = remove_null_values(data)
            if headers.get('content-type') is None:
                headers.update({'content-type': 'application/json'})
            data = json_import.dumps(data).encode('utf-8')
        request['data'] = data

        self.authenticator.authenticate(request)

        # Compress the request body if applicable
        if self.get_enable_gzip_compression() and 'content-encoding' not in headers and request['data'] is not None:
            headers['content-encoding'] = 'gzip'
            request['headers'] = headers
            # If the provided data is a file-like object, we create `GzipStream` which will handle
            # the compression on-the-fly when the requests package starts reading its content.
            # This helps avoid OOM errors when the opened file is too big.
            # In any other cases, we use the in memory compression directly from
            # the `gzip` package for backward compatibility.
            raw_data = request['data']
            request['data'] = GzipStream(raw_data) if isinstance(raw_data, io.IOBase) else gzip.compress(raw_data)

        # Next, we need to process the 'files' argument to try to fill in
        # any missing filenames where possible.
        # 'files' can be a dictionary (i.e { '<part-name>': (<tuple>)} )
        # or a list of tuples [ (<part-name>, (<tuple>))... ]
        # If 'files' is a dictionary we'll convert it to a list of tuples.
        new_files = []
        if files is not None:
            # If 'files' is a dictionary, transform it into a list of tuples.
            if isinstance(files, dict):
                files = remove_null_values(files)
                files = files.items()
            # Next, fill in any missing filenames from file tuples.
            for part_name, file_tuple in files:
                if file_tuple and len(file_tuple) == 3 and file_tuple[0] is None:
                    file = file_tuple[1]
                    if file and hasattr(file, 'name'):
                        filename = basename(file.name)
                        file_tuple = (filename, file_tuple[1], file_tuple[2])
                new_files.append((part_name, file_tuple))
        request['files'] = new_files
        logger.debug('Prepared request [%s %s]', request['method'], request['url'])
        return request

    @staticmethod
    def encode_path_vars(*args: str) -> List[str]:
        """Encode path variables to be substituted into a URL path.

        Arguments:
            args: A list of strings to be URL path encoded

        Returns:
            A list of encoded strings that are safe to substitute into a URL path.
        """
        return (requests.utils.quote(x, safe='') for x in args)

    # The methods below are kept for compatibility and should be removed
    # in the next major release.

    # pylint: disable=protected-access

    @staticmethod
    def _convert_model(val: str) -> None:
        if isinstance(val, str):
            val = json_import.loads(val)
        if hasattr(val, "_to_dict"):
            return val._to_dict()
        return val

    @staticmethod
    def _convert_list(val: list) -> None:
        if isinstance(val, list):
            return ",".join(val)
        return val

    @staticmethod
    def _encode_path_vars(*args) -> None:
        return BaseService.encode_path_vars(*args)



# coding=utf-8
import json
import logging

import requests
import responses

from ibm_cloud_sdk_core import ApiException
from .utils.logger_utils import setup_test_logger

setup_test_logger(logging.ERROR)


@responses.activate
def test_api_exception():
    """Test APIException class"""
    responses.add(
        responses.GET,
        'https://test.com',
        status=500,
        body=json.dumps({'error': 'sorry', 'msg': 'serious error'}),
        content_type='application/json',
    )

    mock_response = requests.get('https://test.com', timeout=None)
    exception = ApiException(500, http_response=mock_response)
    assert exception is not None
    assert exception.message == 'sorry'

    responses.add(
        responses.GET,
        'https://test-again.com',
        status=500,
        body=json.dumps(
            {
                "errors": [
                    {
                        "message": "sorry again",
                    }
                ],
            }
        ),
        content_type='application/json',
    )
    mock_response = requests.get('https://test-again.com', timeout=None)
    exception = ApiException(500, http_response=mock_response)
    assert exception.message == 'sorry again'

    responses.add(
        responses.GET,
        'https://test-once-more.com',
        status=500,
        body=json.dumps({'message': 'sorry once more'}),
        content_type='application/json',
    )
    mock_response = requests.get('https://test-once-more.com', timeout=None)
    exception = ApiException(500, http_response=mock_response)
    assert exception.message == 'sorry once more'

    responses.add(
        responses.GET,
        'https://test-msg.com',
        status=500,
        body=json.dumps({'msg': 'serious error'}),
        content_type='application/json',
    )
    mock_response = requests.get('https://test-msg.com', timeout=None)
    exception = ApiException(500, http_response=mock_response)
    assert exception.message == 'Internal Server Error'

    responses.add(
        responses.GET,
        'https://test-errormessage.com',
        status=500,
        body=json.dumps({'errorMessage': 'IAM error message'}),
        content_type='application/json',
    )
    mock_response = requests.get('https://test-errormessage.com', timeout=None)
    exception = ApiException(500, http_response=mock_response)
    assert exception.message == 'IAM error message'
    assert exception.http_response.text == '{"errorMessage": "IAM error message"}'
    assert exception.http_response.content == b'{"errorMessage": "IAM error message"}'
    assert exception.http_response.json() == {'errorMessage': 'IAM error message'}

    responses.add(
        responses.GET,
        'https://test-for-text.com',
        status=500,
        headers={'X-Global-Transaction-ID': 'xx'},
        body="plain text error",
    )
    mock_response = requests.get('https://test-for-text.com', timeout=None)
    exception = ApiException(500, http_response=mock_response)
    assert exception.message == 'plain text error'
    assert str(exception) == 'Error: plain text error, Status code: 500 , X-global-transaction-id: xx'
    assert exception.http_response.text == 'plain text error'
    assert exception.http_response.content == b'plain text error'



# pylint: disable=missing-docstring
import logging
import os
from ssl import PROTOCOL_TLSv1_1, PROTOCOL_TLSv1_2

import pytest
from requests import exceptions

from ibm_cloud_sdk_core.base_service import BaseService
from ibm_cloud_sdk_core.authenticators import NoAuthAuthenticator
from .utils.logger_utils import setup_test_logger
from .utils.http_utils import local_server


setup_test_logger(logging.ERROR)


# The certificate files that are used in this tests are generated by this command:
# pylint: disable=line-too-long,pointless-string-statement
"""
openssl req -x509 -out test_ssl.crt -keyout test_ssl.key \
  -newkey rsa:2048 -nodes -sha256 -days 36500 \
  -subj '/CN=localhost' -extensions EXT -config <( \
   printf "[dn]\nCN=localhost\n[req]\ndistinguished_name = dn\n[EXT]\nsubjectAltName=DNS:localhost\nkeyUsage=digitalSignature\nextendedKeyUsage=serverAuth")
"""


# Load the certificate and the key files.
cert = os.path.join(os.path.dirname(__file__), '../resources/test_ssl.crt')
key = os.path.join(os.path.dirname(__file__), '../resources/test_ssl.key')


@local_server(3333, PROTOCOL_TLSv1_1, cert, key)
def test_tls_v1_1():
    service = BaseService(service_url='https://localhost:3333', authenticator=NoAuthAuthenticator())
    prepped = service.prepare_request('GET', url='/')
    # The following request should fail, because the server will try
    # to use TLS v1.1 but that's not allowed in our client.
    with pytest.raises(Exception) as exception:
        service.send(prepped, verify=cert)
    # Errors can be differ based on the Python version.
    assert exception.type is exceptions.SSLError or exception.type is exceptions.ConnectionError


@local_server(3334, PROTOCOL_TLSv1_2, cert, key)
def test_tls_v1_2():
    service = BaseService(service_url='https://localhost:3334', authenticator=NoAuthAuthenticator())

    # First call the server with the default configuration.
    # It should fail due to the self-signed SSL cert.
    assert service.disable_ssl_verification is False
    prepped = service.prepare_request('GET', url='/')
    with pytest.raises(exceptions.SSLError, match='certificate verify failed: self-signed certificate'):
        res = service.send(prepped)

    # Next configure it to validate by using our local certificate. Should raise no exception.
    res = service.send(prepped, verify=cert)
    assert res is not None

    # Now disable the SSL verification. The request shouldn't raise any issue.
    service.set_disable_ssl_verification(True)
    assert service.disable_ssl_verification is True
    prepped = service.prepare_request('GET', url='/')
    res = service.send(prepped)
    assert res is not None

    # Lastly, try with an external URL.
    # This test case is mainly here to reproduce the regression
    # in the `requests` package that was introduced in `2.32.3`.
    # More details on the issue can be found here: https://github.com/psf/requests/issues/6730
    service = BaseService(service_url='https://cloud.ibm.com', authenticator=NoAuthAuthenticator())
    assert service.disable_ssl_verification is False

    ssl_context = service.http_adapter.poolmanager.connection_pool_kw.get("ssl_context")
    assert ssl_context is not None
    assert len(ssl_context.get_ca_certs()) > 0

    prepped = service.prepare_request('GET', url='/status')
    res = service.send(prepped)
    assert res is not None



# coding: utf-8

# Copyright 2019, 2024 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=missing-docstring,protected-access,too-few-public-methods,too-many-lines

import gzip
import json
import logging
import os
import ssl
import tempfile
import time
from shutil import copyfile
from typing import Optional
from urllib3.exceptions import ConnectTimeoutError, MaxRetryError

import jwt
import pytest
import responses
import requests

from ibm_cloud_sdk_core import ApiException
from ibm_cloud_sdk_core import BaseService, DetailedResponse
from ibm_cloud_sdk_core import CP4DTokenManager
from ibm_cloud_sdk_core import get_authenticator_from_environment
from ibm_cloud_sdk_core.authenticators import (
    IAMAuthenticator,
    NoAuthAuthenticator,
    Authenticator,
    BasicAuthenticator,
    CloudPakForDataAuthenticator,
)
from .utils.logger_utils import setup_test_logger

# Note that the test_reserved_keys method below will fail if the log level is not WARNING.
setup_test_logger(logging.WARNING)


class IncludeExternalConfigService(BaseService):
    default_service_url = 'https://servicesthatincludeexternalconfig.com/api'

    def __init__(
        self, api_version: str, authenticator: Optional[Authenticator] = None, trace_id: Optional[str] = None
    ) -> None:
        BaseService.__init__(
            self, service_url=self.default_service_url, authenticator=authenticator, disable_ssl_verification=False
        )
        self.api_version = api_version
        self.trace_id = trace_id
        self.configure_service('include-external-config')


class AnyServiceV1(BaseService):
    default_url = 'https://gateway.watsonplatform.net/test/api'

    def __init__(
        self,
        version: str,
        service_url: str = default_url,
        authenticator: Optional[Authenticator] = None,
        disable_ssl_verification: bool = False,
    ) -> None:
        BaseService.__init__(
            self,
            service_url=service_url,
            authenticator=authenticator,
            disable_ssl_verification=disable_ssl_verification,
        )
        self.version = version

    def op_with_path_params(self, path0: str, path1: str) -> DetailedResponse:
        if path0 is None:
            raise ValueError('path0 must be provided')
        if path1 is None:
            raise ValueError('path1 must be provided')
        params = {'version': self.version}
        url = '/v1/foo/{0}/bar/{1}/baz'.format(*self._encode_path_vars(path0, path1))
        request = self.prepare_request(method='GET', url=url, params=params)
        response = self.send(request)
        return response

    def with_http_config(self, http_config: dict) -> DetailedResponse:
        self.set_http_config(http_config)
        request = self.prepare_request(method='GET', url='')
        response = self.send(request)
        return response

    def any_service_call(self) -> DetailedResponse:
        request = self.prepare_request(method='GET', url='')
        response = self.send(request)
        return response

    def head_request(self) -> DetailedResponse:
        request = self.prepare_request(method='HEAD', url='')
        response = self.send(request)
        return response

    def get_document_as_stream(self) -> DetailedResponse:
        params = {'version': self.version}
        url = '/v1/streamjson'
        request = self.prepare_request(method='GET', url=url, params=params)
        response = self.send(request, stream=True)
        return response


def get_access_token() -> str:
    access_token_layout = {
        "username": "dummy",
        "role": "Admin",
        "permissions": ["administrator", "manage_catalog"],
        "sub": "admin",
        "iss": "sss",
        "aud": "sss",
        "uid": "sss",
        "iat": 3600,
        "exp": int(time.time()),
    }

    access_token = jwt.encode(
        access_token_layout, 'secret', algorithm='HS256', headers={'kid': '230498151c214b788dd97f22b85410a5'}
    )
    return access_token


def test_invalid_authenticator():
    with pytest.raises(ValueError) as err:
        AnyServiceV1('2021-08-18')

    assert str(err.value) == 'authenticator must be provided'


@responses.activate
def test_url_encoding():
    service = AnyServiceV1('2017-07-07', authenticator=NoAuthAuthenticator())

    # All characters in path0 _must_ be encoded in path segments
    path0 = ' \"<>^`{}|/\\?#%[]'
    path0_encoded = '%20%22%3C%3E%5E%60%7B%7D%7C%2F%5C%3F%23%25%5B%5D'
    # All non-ASCII chars _must_ be encoded in path segments
    path1 = '比萨浇头'.encode('utf8')  # "pizza toppings"
    path1_encoded = '%E6%AF%94%E8%90%A8%E6%B5%87%E5%A4%B4'

    path_encoded = '/v1/foo/' + path0_encoded + '/bar/' + path1_encoded + '/baz'
    test_url = service.default_url + path_encoded

    responses.add(
        responses.GET, test_url, status=200, body=json.dumps({"foobar": "baz"}), content_type='application/json'
    )

    # Set Host as a default header on the service.
    service.set_default_headers({'Host': 'alternatehost.ibm.com:443'})

    response = service.op_with_path_params(path0, path1)

    assert response is not None
    assert len(responses.calls) == 1
    assert path_encoded in responses.calls[0].request.url
    assert 'version=2017-07-07' in responses.calls[0].request.url

    # Verify that the Host header was set in the request.
    assert responses.calls[0].request.headers.get('Host') == 'alternatehost.ibm.com:443'


@responses.activate
def test_stream_json_response():
    service = AnyServiceV1('2017-07-07', authenticator=NoAuthAuthenticator())

    path = '/v1/streamjson'
    test_url = service.default_url + path

    expected_response = json.dumps({"id": 1, "rev": "v1", "content": "this is a document"})

    # print("Expected response: ", expected_response)

    # Simulate a JSON response
    responses.add(responses.GET, test_url, status=200, body=expected_response, content_type='application/json')

    # Invoke the operation and receive an "iterable" as the response
    response = service.get_document_as_stream()

    assert response is not None
    assert len(responses.calls) == 1

    # retrieve the requests.Response object from the DetailedResponse
    resp = response.get_result()
    assert isinstance(resp, requests.Response)
    assert hasattr(resp, "iter_content")

    # Retrieve the response body, one chunk at a time.
    actual_response = ''
    for chunk in resp.iter_content(chunk_size=3):
        actual_response += chunk.decode("utf-8")

    # print("Actual response: ", actual_response)
    assert actual_response == expected_response


@responses.activate
def test_http_config():
    service = AnyServiceV1('2017-07-07', authenticator=NoAuthAuthenticator())
    responses.add(
        responses.GET,
        service.default_url,
        status=200,
        body=json.dumps({"foobar": "baz"}),
        content_type='application/json',
    )

    response = service.with_http_config({'timeout': 100})
    assert response is not None
    assert len(responses.calls) == 1


def test_fail_http_config():
    service = AnyServiceV1('2017-07-07', authenticator=NoAuthAuthenticator())
    with pytest.raises(TypeError):
        service.with_http_config(None)


@responses.activate
def test_cwd():
    file_path = os.path.join(os.path.dirname(__file__), '../resources/ibm-credentials.env')
    # Try changing working directories to test getting creds from cwd
    cwd = os.getcwd()
    os.chdir(os.path.dirname(file_path))
    iam_authenticator = get_authenticator_from_environment('ibm_watson')
    service = AnyServiceV1('2017-07-07', authenticator=iam_authenticator)
    service.configure_service('ibm_watson')
    os.chdir(cwd)
    assert service.service_url == 'https://cwdserviceurl'
    assert service.authenticator is not None

    # Copy credentials file to cwd to test loading from current working directory
    temp_env_path = os.getcwd() + '/ibm-credentials.env'
    copyfile(file_path, temp_env_path)
    iam_authenticator = get_authenticator_from_environment('ibm_watson')
    service = AnyServiceV1('2017-07-07', authenticator=iam_authenticator)
    service.configure_service('ibm_watson')
    os.remove(temp_env_path)
    assert service.service_url == 'https://cwdserviceurl'
    assert service.authenticator is not None


@responses.activate
def test_iam():
    file_path = os.path.join(os.path.dirname(__file__), '../resources/ibm-credentials-iam.env')
    os.environ['IBM_CREDENTIALS_FILE'] = file_path
    iam_authenticator = get_authenticator_from_environment('ibm-watson')
    service = AnyServiceV1('2017-07-07', authenticator=iam_authenticator)
    assert service.service_url == 'https://gateway.watsonplatform.net/test/api'
    del os.environ['IBM_CREDENTIALS_FILE']
    assert service.authenticator is not None

    response = {
        "access_token": get_access_token(),
        "token_type": "Bearer",
        "expires_in": 3600,
        "expiration": int(time.time()),
        "refresh_token": "jy4gl91BQ",
    }
    responses.add(responses.POST, url='https://iam.cloud.ibm.com/identity/token', body=json.dumps(response), status=200)
    responses.add(
        responses.GET,
        url='https://gateway.watsonplatform.net/test/api',
        body=json.dumps({"foobar": "baz"}),
        content_type='application/json',
    )
    service.any_service_call()
    assert "grant-type%3Aapikey" in responses.calls[0].request.body


def test_no_auth():
    class MadeUp:
        def __init__(self):
            self.lazy = 'made up'

    with pytest.raises(ValueError) as err:
        service = AnyServiceV1('2017-07-07', authenticator=MadeUp())
        service.prepare_request(
            responses.GET,
            url='https://gateway.watsonplatform.net/test/api',
        )
        assert str(err.value) == 'authenticator should be of type Authenticator'

    service = AnyServiceV1('2017-07-07', authenticator=NoAuthAuthenticator())
    service.prepare_request(
        responses.GET,
        url='https://gateway.watsonplatform.net/test/api',
    )
    assert service.authenticator is not None
    assert isinstance(service.authenticator, Authenticator)


def test_for_cp4d():
    cp4d_authenticator = CloudPakForDataAuthenticator('my_username', 'my_password', 'my_url')
    service = AnyServiceV1('2017-07-07', authenticator=cp4d_authenticator)
    assert service.authenticator.token_manager is not None
    assert service.authenticator.token_manager.username == 'my_username'
    assert service.authenticator.token_manager.password == 'my_password'
    assert service.authenticator.token_manager.url == 'my_url/v1/authorize'
    assert isinstance(service.authenticator.token_manager, CP4DTokenManager)


def test_disable_ssl_verification():
    service1 = AnyServiceV1('2017-07-07', authenticator=NoAuthAuthenticator(), disable_ssl_verification=True)
    assert service1.disable_ssl_verification is True

    service1.set_disable_ssl_verification(False)
    assert service1.disable_ssl_verification is False

    cp4d_authenticator = CloudPakForDataAuthenticator('my_username', 'my_password', 'my_url')
    service2 = AnyServiceV1('2017-07-07', authenticator=cp4d_authenticator)
    assert service2.disable_ssl_verification is False
    cp4d_authenticator.set_disable_ssl_verification(True)
    assert service2.authenticator.token_manager.disable_ssl_verification is True


@responses.activate
def test_http_head():
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
    expected_headers = {'Test-Header1': 'value1', 'Test-Header2': 'value2'}
    responses.add(responses.HEAD, service.default_url, status=200, headers=expected_headers, content_type=None)

    response = service.head_request()
    assert response is not None
    assert len(responses.calls) == 1
    assert response.headers is not None
    assert response.headers == expected_headers


@responses.activate
def test_response_with_no_body():
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
    responses.add(responses.GET, service.default_url, status=200, body=None)

    response = service.any_service_call()
    assert response is not None
    assert len(responses.calls) == 1
    assert response.get_result() is None


def test_has_bad_first_or_last_char():
    with pytest.raises(ValueError) as err:
        basic_authenticator = BasicAuthenticator('{my_username}', 'my_password')
        AnyServiceV1('2018-11-20', authenticator=basic_authenticator).prepare_request(
            responses.GET, 'https://gateway.watsonplatform.net/test/api'
        )
    assert (
        str(err.value) == 'The username and password shouldn\'t start or end with curly brackets or quotes. '
        'Please remove any surrounding {, }, or \" characters.'
    )


@responses.activate
def test_request_server_error():
    with pytest.raises(ApiException, match=r'internal server error') as err:
        responses.add(
            responses.GET,
            'https://gateway.watsonplatform.net/test/api',
            status=500,
            body=json.dumps({'error': 'internal server error'}),
            content_type='application/json',
        )
        service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
        prepped = service.prepare_request('GET', url='')
        service.send(prepped)
    assert err.value.code == 500
    assert err.value.status_code == 500
    assert err.value.http_response.headers['Content-Type'] == 'application/json'
    assert err.value.message == 'internal server error'


@responses.activate
def test_request_success_json():
    responses.add(
        responses.GET,
        'https://gateway.watsonplatform.net/test/api',
        status=200,
        body=json.dumps({'foo': 'bar'}),
        content_type='application/json',
    )
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
    prepped = service.prepare_request('GET', url='')
    detailed_response = service.send(prepped)
    assert detailed_response.get_result() == {'foo': 'bar'}

    service = AnyServiceV1('2018-11-20', authenticator=BasicAuthenticator('my_username', 'my_password'))
    service.set_default_headers({'test': 'header'})
    service.set_disable_ssl_verification(True)
    prepped = service.prepare_request('GET', url='')
    detailed_response = service.send(prepped)
    assert detailed_response.get_result() == {'foo': 'bar'}


@responses.activate
def test_request_success_invalid_json():
    # expect an ApiException with JSONDecodeError as the cause when a "success"
    # response contains invalid JSON in response body.
    with pytest.raises(ApiException, match=r'Error processing the HTTP response') as err:
        responses.add(
            responses.GET,
            'https://gateway.watsonplatform.net/test/api',
            status=200,
            body='{ "invalid": "json", "response"',
            content_type='application/json; charset=utf8',
        )
        service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
        prepped = service.prepare_request('GET', url='')
        service.send(prepped)
    assert err.value.code == 200
    assert err.value.status_code == 200
    assert err.value.http_response.headers['Content-Type'] == 'application/json; charset=utf8'
    assert isinstance(err.value.__cause__, requests.exceptions.JSONDecodeError)
    assert "Expecting ':' delimiter: line 1" in str(err.value.__cause__)


@responses.activate
def test_request_success_response():
    expected_body = '{"foo": "bar", "description": "this\nis\na\ndescription"}'
    responses.add(
        responses.GET,
        'https://gateway.watsonplatform.net/test/api',
        status=200,
        body=expected_body,
        content_type='application/json',
    )
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
    prepped = service.prepare_request('GET', url='')
    detailed_response = service.send(prepped)
    assert detailed_response.get_result() == {"foo": "bar", "description": "this\nis\na\ndescription"}


@responses.activate
def test_request_success_nonjson():
    responses.add(
        responses.GET,
        'https://gateway.watsonplatform.net/test/api',
        status=200,
        body='<h1>Hola, amigo!</h1>',
        content_type='text/html',
    )
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
    prepped = service.prepare_request('GET', url='')
    detailed_response = service.send(prepped)
    # It's odd that we have to call ".text" to get the string value
    # (see issue 3557)
    assert detailed_response.get_result().text == '<h1>Hola, amigo!</h1>'


@responses.activate
def test_request_fail_401_nonerror_json():
    # response body not an error object, so we expect the default error message.
    error_msg = 'Unauthorized: Access is denied due to invalid credentials'
    with pytest.raises(ApiException, match=error_msg) as err:
        responses.add(
            responses.GET,
            'https://gateway.watsonplatform.net/test/api',
            status=401,
            body=json.dumps({'foo': 'bar'}),
            content_type='application/json',
        )
        service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
        prepped = service.prepare_request('GET', url='')
        service.send(prepped)
    assert err.value.code == 401
    assert err.value.status_code == 401
    assert err.value.http_response.headers['Content-Type'] == 'application/json'
    assert err.value.message == error_msg


@responses.activate
def test_request_fail_401_error_json():
    # response body is an error object, so we expect to get the message from there.
    error_msg = 'You dont need to know...'
    with pytest.raises(ApiException, match=error_msg) as err:
        responses.add(
            responses.GET,
            'https://gateway.watsonplatform.net/test/api',
            status=401,
            body=json.dumps({'message': error_msg}),
            content_type='application/json',
        )
        service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
        prepped = service.prepare_request('GET', url='')
        service.send(prepped)
    assert err.value.code == 401
    assert err.value.status_code == 401
    assert err.value.http_response.headers['Content-Type'] == 'application/json'
    assert err.value.message == error_msg


@responses.activate
def test_request_fail_401_nonjson():
    response_body = 'You dont have a need to know...'
    with pytest.raises(ApiException, match=response_body) as err:
        responses.add(
            responses.GET,
            'https://gateway.watsonplatform.net/test/api',
            status=401,
            body=response_body,
            content_type='text/plain',
        )
        service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
        prepped = service.prepare_request('GET', url='')
        service.send(prepped)
    assert err.value.code == 401
    assert err.value.status_code == 401
    assert err.value.http_response.headers['Content-Type'] == 'text/plain'
    assert err.value.message == response_body


@responses.activate
def test_request_fail_401_badjson():
    # if an error response contains invalid JSON, then we should
    # end up with 'Unknown error' as the message since we couldn't get
    # the actual error message from the response body.
    response_body = 'This is not a JSON object'
    with pytest.raises(ApiException, match=response_body) as err:
        responses.add(
            responses.GET,
            'https://gateway.watsonplatform.net/test/api',
            status=401,
            body=response_body,
            content_type='application/json',
        )
        service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
        prepped = service.prepare_request('GET', url='')
        service.send(prepped)
    assert err.value.code == 401
    assert err.value.status_code == 401
    assert err.value.http_response.headers['Content-Type'] == 'application/json'
    assert err.value.message == response_body


def test_misc_methods():
    class MockModel:
        def __init__(self, xyz=None):
            self.xyz = xyz

        def _to_dict(self):
            _dict = {}
            if hasattr(self, 'xyz') and self.xyz is not None:
                _dict['xyz'] = self.xyz
            return _dict

        @classmethod
        def _from_dict(cls, _dict):
            args = {}
            if 'xyz' in _dict:
                args['xyz'] = _dict.get('xyz')
            return cls(**args)

    mock = MockModel('foo')
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
    model1 = service._convert_model(mock)
    assert model1 == {'xyz': 'foo'}

    model2 = service._convert_model("{\"xyz\": \"foo\"}")
    assert model2 is not None
    assert model2['xyz'] == 'foo'

    temp = ['default', '123']
    res_str = service._convert_list(temp)
    assert res_str == 'default,123'

    temp2 = 'default123'
    res_str2 = service._convert_list(temp2)
    assert res_str2 == temp2


def test_default_headers():
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
    service.set_default_headers({'xxx': 'yyy'})
    assert service.default_headers == {'xxx': 'yyy'}
    with pytest.raises(TypeError):
        service.set_default_headers('xxx')


def test_set_service_url():
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
    with pytest.raises(ValueError) as err:
        service.set_service_url('{url}')
    assert (
        str(err.value) == 'The service url shouldn\'t start or end with curly brackets or quotes. '
        'Be sure to remove any {} and \" characters surrounding your service url'
    )

    service.set_service_url('my_url')


def test_http_client():
    auth = BasicAuthenticator('my_username', 'my_password')
    service = AnyServiceV1('2018-11-20', authenticator=auth)
    assert isinstance(service.get_http_client(), requests.sessions.Session)
    assert service.get_http_client().headers.get('Accept-Encoding') == 'gzip, deflate'

    new_http_client = requests.Session()
    new_http_client.headers.update({'Accept-Encoding': 'gzip'})
    service.set_http_client(http_client=new_http_client)
    assert service.get_http_client().headers.get('Accept-Encoding') == 'gzip'

    with pytest.raises(TypeError):
        service.set_http_client("bad_argument_type")


def test_get_authenticator():
    auth = BasicAuthenticator('my_username', 'my_password')
    service = AnyServiceV1('2018-11-20', authenticator=auth)
    assert service.get_authenticator() is not None


def test_gzip_compression():
    # Should return uncompressed data when gzip is off
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
    assert not service.get_enable_gzip_compression()
    prepped = service.prepare_request('GET', url='', data=json.dumps({"foo": "bar"}))
    assert prepped['data'] == b'{"foo": "bar"}'
    assert prepped['headers'].get('content-encoding') != 'gzip'

    # Should return compressed data when gzip is on
    service.set_enable_gzip_compression(True)
    assert service.get_enable_gzip_compression()
    prepped = service.prepare_request('GET', url='', data=json.dumps({"foo": "bar"}))
    assert prepped['data'] == gzip.compress(b'{"foo": "bar"}')
    assert prepped['headers'].get('content-encoding') == 'gzip'

    # Should return compressed data when gzip is on for non-json data
    assert service.get_enable_gzip_compression()
    prepped = service.prepare_request('GET', url='', data=b'rawdata')
    assert prepped['data'] == gzip.compress(b'rawdata')
    assert prepped['headers'].get('content-encoding') == 'gzip'

    # Should return compressed data when gzip is on for gzip file data
    assert service.get_enable_gzip_compression()
    with tempfile.TemporaryFile(mode='w+b') as t_f:
        with gzip.GzipFile(mode='wb', fileobj=t_f) as gz_f:
            gz_f.write(json.dumps({"foo": "bar"}).encode())
        with gzip.GzipFile(mode='rb', fileobj=t_f) as gz_f:
            gzip_data = gz_f.read()
        prepped = service.prepare_request('GET', url='', data=gzip_data)
        assert prepped['data'] == gzip.compress(t_f.read())
        assert prepped['headers'].get('content-encoding') == 'gzip'

    # Should return compressed json data when gzip is on for gzip file json data
    assert service.get_enable_gzip_compression()
    with tempfile.TemporaryFile(mode='w+b') as t_f:
        with gzip.GzipFile(mode='wb', fileobj=t_f) as gz_f:
            gz_f.write("rawdata".encode())
        with gzip.GzipFile(mode='rb', fileobj=t_f) as gz_f:
            gzip_data = gz_f.read()
        prepped = service.prepare_request('GET', url='', data=gzip_data)
        assert prepped['data'] == gzip.compress(t_f.read())
        assert prepped['headers'].get('content-encoding') == 'gzip'

    # Should return uncompressed data when content-encoding is set
    assert service.get_enable_gzip_compression()
    prepped = service.prepare_request(
        'GET', url='', headers={"content-encoding": "gzip"}, data=json.dumps({"foo": "bar"})
    )
    assert prepped['data'] == b'{"foo": "bar"}'
    assert prepped['headers'].get('content-encoding') == 'gzip'


def test_gzip_compression_file_input():
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
    service.set_enable_gzip_compression(True)

    # Should return file-like object with the compressed data when compression is on
    # and the input is a file, opened for reading in binary mode.
    raw_data = b'rawdata'
    with tempfile.TemporaryFile(mode='w+b') as tmp_file:
        tmp_file.write(raw_data)
        tmp_file.seek(0)

        prepped = service.prepare_request('GET', url='', data=tmp_file)
        assert prepped['data'].read() == gzip.compress(raw_data)
        assert prepped['headers'].get('content-encoding') == 'gzip'
        assert prepped['data'].read() == b''

    # Simulate the requests (urllib3) package reading method for binary files.
    with tempfile.TemporaryFile(mode='w+b') as tmp_file:
        tmp_file.write(raw_data)
        tmp_file.seek(0)

        prepped = service.prepare_request('GET', url='', data=tmp_file)
        compressed = b''
        for chunk in prepped['data']:
            compressed += chunk

        assert compressed == gzip.compress(raw_data)

    # Make sure the decompression works fine.
    assert gzip.decompress(compressed) == raw_data

    # Should return file-like object with the compressed data when compression is on
    # and the input is a file, opened for reading in text mode.
    assert service.get_enable_gzip_compression()
    text_data = 'textdata'
    with tempfile.TemporaryFile(mode='w+') as tmp_file:
        tmp_file.write(text_data)
        tmp_file.seek(0)

        prepped = service.prepare_request('GET', url='', data=tmp_file)
        assert prepped['data'].read() == gzip.compress(text_data.encode())
        assert prepped['headers'].get('content-encoding') == 'gzip'
        assert prepped['data'].read() == b''

    # Simulate the requests (urllib3) package reading method for text files.
    with tempfile.TemporaryFile(mode='w+') as tmp_file:
        tmp_file.write(text_data)
        tmp_file.seek(0)

        prepped = service.prepare_request('GET', url='', data=tmp_file)
        compressed = b''
        for chunk in prepped['data']:
            compressed += chunk

        assert compressed == gzip.compress(text_data.encode())

    # Make sure the decompression works fine.
    assert gzip.decompress(compressed).decode() == text_data


def test_gzip_compression_external():
    # Should set gzip compression from external config
    file_path = os.path.join(os.path.dirname(__file__), '../resources/ibm-credentials-gzip.env')
    os.environ['IBM_CREDENTIALS_FILE'] = file_path
    service = IncludeExternalConfigService('v1', authenticator=NoAuthAuthenticator())
    assert service.service_url == 'https://mockurl'
    assert service.get_enable_gzip_compression() is True
    prepped = service.prepare_request('GET', url='', data=json.dumps({"foo": "bar"}))
    assert prepped['data'] == gzip.compress(b'{"foo": "bar"}')
    assert prepped['headers'].get('content-encoding') == 'gzip'


def test_retry_config_default():
    service = BaseService(service_url='https://mockurl/', authenticator=NoAuthAuthenticator())
    service.enable_retries()
    assert service.retry_config.total == 4
    assert service.retry_config.backoff_factor == 1.0
    assert service.retry_config.backoff_max == 30.0
    assert service.http_client.get_adapter('https://').max_retries.total == 4

    # Ensure retries fail after 4 retries
    error = ConnectTimeoutError()
    retry = service.http_client.get_adapter('https://').max_retries
    retry = retry.increment(error=error)
    retry = retry.increment(error=error)
    retry = retry.increment(error=error)
    retry = retry.increment(error=error)
    with pytest.raises(MaxRetryError) as retry_err:
        retry.increment(error=error)
    assert retry_err.value.reason == error


def test_retry_config_disable():
    # Test disabling retries
    service = BaseService(service_url='https://mockurl/', authenticator=NoAuthAuthenticator())
    service.enable_retries()
    service.disable_retries()
    assert service.retry_config is None
    assert service.http_client.get_adapter('https://').max_retries.total == 0

    # Ensure retries are not started after one connection attempt
    error = ConnectTimeoutError()
    retry = service.http_client.get_adapter('https://').max_retries
    with pytest.raises(MaxRetryError) as retry_err:
        retry.increment(error=error)
    assert retry_err.value.reason == error


def test_retry_config_non_default():
    service = BaseService(service_url='https://mockurl/', authenticator=NoAuthAuthenticator())
    service.enable_retries(2, 10.0)
    assert service.retry_config.total == 2
    assert service.retry_config.backoff_factor == 1.0
    assert service.retry_config.backoff_max == 10.0

    # Ensure retries fail after 2 retries
    error = ConnectTimeoutError()
    retry = service.http_client.get_adapter('https://').max_retries
    retry = retry.increment(error=error)
    retry = retry.increment(error=error)
    with pytest.raises(MaxRetryError) as retry_err:
        retry.increment(error=error)
    assert retry_err.value.reason == error


def test_retry_config_external():
    file_path = os.path.join(os.path.dirname(__file__), '../resources/ibm-credentials-retry.env')
    os.environ['IBM_CREDENTIALS_FILE'] = file_path
    service = IncludeExternalConfigService('v1', authenticator=NoAuthAuthenticator())
    assert service.retry_config.total == 3
    assert service.retry_config.backoff_factor == 1.0
    assert service.retry_config.backoff_max == 25.0

    # Ensure retries fail after 3 retries
    error = ConnectTimeoutError()
    retry = service.http_client.get_adapter('https://').max_retries
    retry = retry.increment(error=error)
    retry = retry.increment(error=error)
    retry = retry.increment(error=error)
    with pytest.raises(MaxRetryError) as retry_err:
        retry.increment(error=error)
    assert retry_err.value.reason == error


@responses.activate
def test_user_agent_header():
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
    user_agent_header = service.user_agent_header
    assert user_agent_header is not None
    assert user_agent_header['User-Agent'].startswith('ibm-python-sdk-core-')

    responses.add(responses.GET, 'https://gateway.watsonplatform.net/test/api', status=200, body='some text')
    prepped = service.prepare_request('GET', url='', headers={'user-agent': 'my_user_agent'})
    response = service.send(prepped)
    assert response.get_result().request.headers.get('user-agent') == 'my_user_agent'

    prepped = service.prepare_request('GET', url='', headers=None)
    response = service.send(prepped)
    assert response.get_result().request.headers.get('user-agent') == user_agent_header['User-Agent']


# Note that this test will fail if the log level is not set to WARNING above.
# This is due to the fact that we are checking that the expected warning messages
# are located in specific locations within the captured logs, and this only works correctly
# if only warning and error messages are being logged.
@responses.activate
def test_reserved_keys(caplog):
    service = AnyServiceV1('2021-07-02', authenticator=NoAuthAuthenticator())
    responses.add(responses.GET, 'https://gateway.watsonplatform.net/test/api', status=200, body='some text')
    prepped = service.prepare_request('GET', url='', headers={'key': 'OK'})
    response = service.send(
        prepped, headers={'key': 'bad'}, method='POST', url='localhost', cookies=None, hooks={'response': []}
    )
    assert response.get_result().request.headers.get('key') == 'OK'
    assert response.get_result().request.url == 'https://gateway.watsonplatform.net/test/api'
    assert response.get_result().request.method == 'GET'
    assert response.get_result().request.hooks == {'response': []}
    assert caplog.record_tuples[0][2] == '"method" has been removed from the request'
    assert caplog.record_tuples[1][2] == '"url" has been removed from the request'
    assert caplog.record_tuples[2][2] == '"cookies" has been removed from the request'


@responses.activate
def test_ssl_error():
    responses.add(responses.GET, 'https://gateway.watsonplatform.net/test/api', body=requests.exceptions.SSLError())
    service = AnyServiceV1('2021-08-18', authenticator=NoAuthAuthenticator())
    with pytest.raises(requests.exceptions.SSLError):
        prepped = service.prepare_request('GET', url='')
        service.send(prepped)


def test_files_dict():
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())

    form_data = {}
    with open(
        os.path.join(os.path.dirname(__file__), '../resources/ibm-credentials-iam.env'), 'r', encoding='utf-8'
    ) as file:
        form_data['file1'] = (None, file, 'application/octet-stream')
    form_data['string1'] = (None, 'hello', 'text/plain')
    request = service.prepare_request('GET', url='', headers={'X-opt-out': True}, files=form_data)
    files = request['files']
    assert isinstance(files, list)
    assert len(files) == 2
    files_dict = dict(files)
    file1 = files_dict['file1']
    assert file1[0] == 'ibm-credentials-iam.env'
    string1 = files_dict['string1']
    assert string1[0] is None


def test_files_list():
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())

    form_data = []
    with open(
        os.path.join(os.path.dirname(__file__), '../resources/ibm-credentials-iam.env'), 'r', encoding='utf-8'
    ) as file:
        form_data.append(('file1', (None, file, 'application/octet-stream')))
    form_data.append(('string1', (None, 'hello', 'text/plain')))
    request = service.prepare_request('GET', url='', headers={'X-opt-out': True}, files=form_data)
    files = request['files']
    assert isinstance(files, list)
    assert len(files) == 2
    files_dict = dict(files)
    file1 = files_dict['file1']
    assert file1[0] == 'ibm-credentials-iam.env'
    string1 = files_dict['string1']
    assert string1[0] is None


def test_files_duplicate_parts():
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())

    form_data = []
    with open(
        os.path.join(os.path.dirname(__file__), '../resources/ibm-credentials-iam.env'), 'r', encoding='utf-8'
    ) as file:
        form_data.append(('creds_file', (None, file, 'application/octet-stream')))
    with open(
        os.path.join(os.path.dirname(__file__), '../resources/ibm-credentials-basic.env'), 'r', encoding='utf-8'
    ) as file:
        form_data.append(('creds_file', (None, file, 'application/octet-stream')))
    with open(
        os.path.join(os.path.dirname(__file__), '../resources/ibm-credentials-bearer.env'), 'r', encoding='utf-8'
    ) as file:
        form_data.append(('creds_file', (None, file, 'application/octet-stream')))
    request = service.prepare_request('GET', url='', headers={'X-opt-out': True}, files=form_data)
    files = request['files']
    assert isinstance(files, list)
    assert len(files) == 3
    for part_name, file_tuple in files:
        assert part_name == 'creds_file'
        assert file_tuple[0] is not None


def test_json():
    service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
    req = service.prepare_request('POST', url='', headers={'X-opt-out': True}, data={'hello': 'world', 'fóó': 'bår'})
    assert req.get('data') == b'{"hello": "world", "f\\u00f3\\u00f3": "b\\u00e5r"}'


def test_service_url_handling():
    service = AnyServiceV1('2018-11-20', service_url='https://host///////', authenticator=NoAuthAuthenticator())
    assert service.service_url == 'https://host'

    service.set_service_url('https://host/')
    assert service.service_url == 'https://host'

    req = service.prepare_request('POST', url='/path/', headers={'X-opt-out': True}, data={'hello': 'world'})
    assert req.get('url') == 'https://host/path/'

    service = AnyServiceV1('2018-11-20', service_url='https://host/', authenticator=NoAuthAuthenticator())
    assert service.service_url == 'https://host'

    service.set_service_url('https://host/')
    assert service.service_url == 'https://host'

    req = service.prepare_request('POST', url='/', headers={'X-opt-out': True}, data={'hello': 'world'})
    assert req.get('url') == 'https://host/'

    req = service.prepare_request('POST', url='////', headers={'X-opt-out': True}, data={'hello': 'world'})
    assert req.get('url') == 'https://host/'

    service.set_service_url(None)
    assert service.service_url is None

    service = AnyServiceV1('2018-11-20', service_url='/', authenticator=NoAuthAuthenticator())
    assert service.service_url == ''

    service.set_service_url('/')
    assert service.service_url == ''

    with pytest.raises(ValueError) as err:
        service.prepare_request('POST', url='/', headers={'X-opt-out': True}, data={'hello': 'world'})
    assert str(err.value) == 'The service_url is required'


def test_service_url_slash():
    service = AnyServiceV1('2018-11-20', service_url='/', authenticator=NoAuthAuthenticator())
    assert service.service_url == ''
    with pytest.raises(ValueError) as err:
        service.prepare_request('POST', url='/', headers={'X-opt-out': True}, data={'hello': 'world'})
    assert str(err.value) == 'The service_url is required'


def test_service_url_not_set():
    service = BaseService(service_url='', authenticator=NoAuthAuthenticator())
    with pytest.raises(ValueError) as err:
        service.prepare_request('POST', url='')
    assert str(err.value) == 'The service_url is required'


def test_setting_proxy():
    service = BaseService(service_url='test', authenticator=IAMAuthenticator('wonder woman'))
    assert service.authenticator is not None
    assert service.authenticator.token_manager.http_config == {}

    http_config = {"proxies": {"http": "user:password@host:port"}}
    service.set_http_config(http_config)
    assert service.authenticator.token_manager.http_config == http_config

    service2 = BaseService(service_url='test', authenticator=BasicAuthenticator('marvellous', 'mrs maisel'))
    service2.set_http_config(http_config)
    assert service2.authenticator is not None


def test_configure_service():
    file_path = os.path.join(os.path.dirname(__file__), '../resources/ibm-credentials-external.env')
    os.environ['IBM_CREDENTIALS_FILE'] = file_path
    service = IncludeExternalConfigService('v1', authenticator=NoAuthAuthenticator())
    assert service.service_url == 'https://externallyconfigured.com/api'
    assert service.disable_ssl_verification is True
    # The authenticator should not be changed as a result of configure_service()
    assert isinstance(service.get_authenticator(), NoAuthAuthenticator)


def test_configure_service_error():
    service = BaseService(service_url='v1', authenticator=NoAuthAuthenticator())
    with pytest.raises(ValueError) as err:
        service.configure_service(None)
    assert str(err.value) == 'Service_name must be of type string.'


def test_min_ssl_version():
    service = AnyServiceV1('2022-03-08', authenticator=NoAuthAuthenticator())
    adapter = service.http_client.get_adapter('https://')
    ssl_context = adapter.poolmanager.connection_pool_kw.get('ssl_context', None)
    assert ssl_context is not None
    assert ssl_context.minimum_version == ssl.TLSVersion.TLSv1_2



# pylint: disable=missing-docstring
import logging
import json
import time
import jwt
import pytest
import responses

from ibm_cloud_sdk_core.authenticators import MCSPV2Authenticator, Authenticator
from ibm_cloud_sdk_core import MCSPV2TokenManager
from .utils.logger_utils import setup_test_logger

setup_test_logger(logging.ERROR)

MOCK_APIKEY = 'my-api-key'
MOCK_URL = 'https://mcspv2.ibm.com'
MOCK_SCOPE_COLLECTION_TYPE = 'accounts'
MOCK_SCOPE_ID = 'global_account'
MOCK_CALLER_EXT_CLAIM = {"productID": "prod-123"}
MOCK_HEADERS = {"header1": "value1", "header2": "value2"}
MOCK_PROXIES = {"https": "proxy1", "http": "proxy2"}
MOCK_PATH = '/api/2.0/{0}/{1}/apikeys/token'.format(MOCK_SCOPE_COLLECTION_TYPE, MOCK_SCOPE_ID)


# pylint: disable=too-many-statements
def test_mcspv2_authenticator1():
    # Use only required properties.
    authenticator = MCSPV2Authenticator(
        apikey=MOCK_APIKEY,
        url=MOCK_URL,
        scope_collection_type=MOCK_SCOPE_COLLECTION_TYPE,
        scope_id=MOCK_SCOPE_ID,
    )
    assert authenticator is not None
    assert authenticator.authentication_type() == Authenticator.AUTHTYPE_MCSPV2
    assert authenticator.token_manager.apikey == MOCK_APIKEY
    assert authenticator.token_manager.url == MOCK_URL
    assert authenticator.token_manager.scope_collection_type == MOCK_SCOPE_COLLECTION_TYPE
    assert authenticator.token_manager.scope_id == MOCK_SCOPE_ID
    assert authenticator.token_manager.include_builtin_actions is False
    assert authenticator.token_manager.include_custom_actions is False
    assert authenticator.token_manager.include_roles is True
    assert authenticator.token_manager.prefix_roles is False
    assert authenticator.token_manager.caller_ext_claim is None
    assert authenticator.token_manager.disable_ssl_verification is False
    assert authenticator.token_manager.headers is None
    assert authenticator.token_manager.proxies is None

    # Test setter functions.
    authenticator.set_scope_collection_type("subscriptions")
    assert authenticator.token_manager.scope_collection_type == "subscriptions"

    with pytest.raises(TypeError) as err:
        authenticator.set_scope_collection_type(None)
    assert str(err.value) == '"scope_collection_type" must be a string'

    authenticator.set_scope_id("new_id")
    assert authenticator.token_manager.scope_id == "new_id"

    with pytest.raises(TypeError) as err:
        authenticator.set_scope_id(None)
    assert str(err.value) == '"scope_id" must be a string'

    authenticator.set_include_builtin_actions(True)
    assert authenticator.token_manager.include_builtin_actions is True

    with pytest.raises(TypeError) as err:
        authenticator.set_include_builtin_actions('True')
    assert str(err.value) == '"include_builtin_actions" must be a bool'

    authenticator.set_include_custom_actions(True)
    assert authenticator.token_manager.include_custom_actions is True

    with pytest.raises(TypeError) as err:
        authenticator.set_include_custom_actions('not a bool')
    assert str(err.value) == '"include_custom_actions" must be a bool'

    authenticator.set_include_roles(True)
    assert authenticator.token_manager.include_roles is True

    with pytest.raises(TypeError) as err:
        authenticator.set_include_roles('nope')
    assert str(err.value) == '"include_roles" must be a bool'

    authenticator.set_prefix_roles(True)
    assert authenticator.token_manager.prefix_roles is True

    with pytest.raises(TypeError) as err:
        authenticator.set_prefix_roles('maybe')
    assert str(err.value) == '"prefix_roles" must be a bool'

    authenticator.set_caller_ext_claim(MOCK_CALLER_EXT_CLAIM)
    assert authenticator.token_manager.caller_ext_claim == MOCK_CALLER_EXT_CLAIM

    with pytest.raises(TypeError) as err:
        authenticator.set_caller_ext_claim('not a dictionary')
    assert str(err.value) == '"caller_ext_claim" must be a dictionary or None'

    authenticator.set_disable_ssl_verification(True)
    assert authenticator.token_manager.disable_ssl_verification is True

    with pytest.raises(TypeError) as err:
        authenticator.set_disable_ssl_verification('not a bool')
    assert str(err.value) == '"disable_ssl_verification" must be a bool'

    authenticator.set_headers(MOCK_HEADERS)
    assert authenticator.token_manager.headers == MOCK_HEADERS

    with pytest.raises(TypeError) as err:
        authenticator.set_headers('not a dictionary')
    assert str(err.value) == '"headers" must be a dictionary or None'

    authenticator.set_proxies(MOCK_PROXIES)
    assert authenticator.token_manager.proxies == MOCK_PROXIES

    with pytest.raises(TypeError) as err:
        authenticator.set_proxies('not a dictionary')
    assert str(err.value) == '"proxies" must be a dictionary or None'


def test_mcspv2_authenticator2():
    # Test with all properties.
    authenticator = MCSPV2Authenticator(
        apikey=MOCK_APIKEY,
        url=MOCK_URL,
        scope_collection_type=MOCK_SCOPE_COLLECTION_TYPE,
        scope_id=MOCK_SCOPE_ID,
        include_builtin_actions=True,
        include_custom_actions=True,
        include_roles=False,
        prefix_roles=True,
        caller_ext_claim=MOCK_CALLER_EXT_CLAIM,
        disable_ssl_verification=True,
        headers=MOCK_HEADERS,
        proxies=MOCK_PROXIES,
    )
    assert authenticator.token_manager.apikey == MOCK_APIKEY
    assert authenticator.token_manager.url == MOCK_URL
    assert authenticator.token_manager.scope_collection_type == MOCK_SCOPE_COLLECTION_TYPE
    assert authenticator.token_manager.scope_id == MOCK_SCOPE_ID
    assert authenticator.token_manager.include_builtin_actions is True
    assert authenticator.token_manager.include_custom_actions is True
    assert authenticator.token_manager.include_roles is False
    assert authenticator.token_manager.prefix_roles is True
    assert authenticator.token_manager.caller_ext_claim == MOCK_CALLER_EXT_CLAIM
    assert authenticator.token_manager.disable_ssl_verification is True
    assert authenticator.token_manager.headers == MOCK_HEADERS
    assert authenticator.token_manager.proxies == MOCK_PROXIES


def test_mcsp_authenticator_validate_failed():

    # Check each property individually.
    with pytest.raises(TypeError) as err:
        MCSPV2Authenticator(
            apikey=None,
            url=MOCK_URL,
            scope_collection_type=MOCK_SCOPE_COLLECTION_TYPE,
            scope_id=MOCK_SCOPE_ID,
        )
    assert str(err.value) == '"apikey" must be a string'

    with pytest.raises(TypeError) as err:
        MCSPV2Authenticator(
            apikey=MOCK_APIKEY,
            url=None,
            scope_collection_type=MOCK_SCOPE_COLLECTION_TYPE,
            scope_id=MOCK_SCOPE_ID,
        )
    assert str(err.value) == '"url" must be a string'

    with pytest.raises(TypeError) as err:
        MCSPV2Authenticator(
            apikey=MOCK_APIKEY,
            url=MOCK_URL,
            scope_collection_type=None,
            scope_id=MOCK_SCOPE_ID,
        )
    assert str(err.value) == '"scope_collection_type" must be a string'

    with pytest.raises(TypeError) as err:
        MCSPV2Authenticator(
            apikey=MOCK_APIKEY,
            url=MOCK_URL,
            scope_collection_type=MOCK_SCOPE_COLLECTION_TYPE,
            scope_id=None,
        )
    assert str(err.value) == '"scope_id" must be a string'

    with pytest.raises(TypeError) as err:
        MCSPV2Authenticator(
            apikey=MOCK_APIKEY,
            url=MOCK_URL,
            scope_collection_type=MOCK_SCOPE_COLLECTION_TYPE,
            scope_id=MOCK_SCOPE_ID,
            include_builtin_actions='not a bool',
        )
    assert str(err.value) == '"include_builtin_actions" must be a bool'

    with pytest.raises(TypeError) as err:
        MCSPV2Authenticator(
            apikey=MOCK_APIKEY,
            url=MOCK_URL,
            scope_collection_type=MOCK_SCOPE_COLLECTION_TYPE,
            scope_id=MOCK_SCOPE_ID,
            include_custom_actions=None,
        )
    assert str(err.value) == '"include_custom_actions" must be a bool'

    with pytest.raises(TypeError) as err:
        MCSPV2Authenticator(
            apikey=MOCK_APIKEY,
            url=MOCK_URL,
            scope_collection_type=MOCK_SCOPE_COLLECTION_TYPE,
            scope_id=MOCK_SCOPE_ID,
            include_roles=382636,
        )
    assert str(err.value) == '"include_roles" must be a bool'

    with pytest.raises(TypeError) as err:
        MCSPV2Authenticator(
            apikey=MOCK_APIKEY,
            url=MOCK_URL,
            scope_collection_type=MOCK_SCOPE_COLLECTION_TYPE,
            scope_id=MOCK_SCOPE_ID,
            prefix_roles=None,
        )
    assert str(err.value) == '"prefix_roles" must be a bool'


# utility function to construct a mock token server response containing an access token.
def get_mock_token_response(issued_at: int, time_to_live: int) -> str:
    access_token_layout = {
        "username": "dummy",
        "role": "Admin",
        "permissions": ["administrator", "manage_catalog"],
        "sub": "admin",
        "iss": "sss",
        "aud": "sss",
        "uid": "sss",
        "iat": issued_at,
        "exp": issued_at + time_to_live,
    }

    access_token = jwt.encode(
        access_token_layout, 'secret', algorithm='HS256', headers={'kid': '230498151c214b788dd97f22b85410a5'}
    )

    token_server_response = {
        "token": access_token,
        "token_type": "Bearer",
        "expires_in": time_to_live,
        "expiration": issued_at + time_to_live,
    }

    # For convenience, return both the server response and the access_token.
    return (json.dumps(token_server_response), access_token)


@responses.activate
def test_get_token():
    (response, access_token) = get_mock_token_response(int(time.time()), 7200)
    responses.add(responses.POST, MOCK_URL + MOCK_PATH, body=response, status=200)

    auth_headers = {'Host': 'mcsp.cloud.ibm.com:443'}
    authenticator = MCSPV2Authenticator(
        apikey=MOCK_APIKEY,
        url=MOCK_URL,
        scope_collection_type=MOCK_SCOPE_COLLECTION_TYPE,
        scope_id=MOCK_SCOPE_ID,
        headers=auth_headers,
    )

    # Authenticate the request and verify the Authorization header.
    request = {'headers': {}}
    authenticator.authenticate(request)
    assert request['headers']['Authorization'] == 'Bearer ' + access_token

    # Verify that the "get token" request contained the Host header.
    assert responses.calls[0].request.headers.get('Host') == 'mcsp.cloud.ibm.com:443'


@responses.activate
def test_get_token_cached():
    (response, access_token) = get_mock_token_response(int(time.time()), 7200)
    responses.add(responses.POST, MOCK_URL + MOCK_PATH, body=response, status=200)

    authenticator = MCSPV2Authenticator(
        apikey=MOCK_APIKEY,
        url=MOCK_URL,
        scope_collection_type=MOCK_SCOPE_COLLECTION_TYPE,
        scope_id=MOCK_SCOPE_ID,
    )

    # Authenticate the request and verify the Authorization header.
    request = {'headers': {}}
    authenticator.authenticate(request)
    assert request['headers']['Authorization'] == 'Bearer ' + access_token

    # Authenticate a second request and verify that we used the same access token.
    request = {'headers': {}}
    authenticator.authenticate(request)
    assert request['headers']['Authorization'] == 'Bearer ' + access_token


@responses.activate
def test_get_token_background_refresh():
    t1 = time.time()
    t2 = t1 + 7200

    # Setup the first token response.
    (response1, access_token1) = get_mock_token_response(int(t1), 7200)
    responses.add(responses.POST, MOCK_URL + MOCK_PATH, body=response1, status=200)

    # Setup the second token response.
    (response2, access_token2) = get_mock_token_response(int(t2), 7200)
    responses.add(responses.POST, MOCK_URL + MOCK_PATH, body=response2, status=200)

    authenticator = MCSPV2Authenticator(
        apikey=MOCK_APIKEY,
        url=MOCK_URL,
        scope_collection_type=MOCK_SCOPE_COLLECTION_TYPE,
        scope_id=MOCK_SCOPE_ID,
    )

    # Authenticate the request and verify that the first access_token is used.
    request = {'headers': {}}
    authenticator.authenticate(request)
    assert request['headers']['Authorization'] == 'Bearer ' + access_token1

    # Now put the token manager in the refresh window to trigger a background refresh scenario.
    authenticator.token_manager.refresh_time = t1 - 1

    # Authenticate a second request and verify that the correct access token is used.
    # Note: Ideally, the token manager would trigger the refresh in a separate thread
    # and it "should" return the first access token for this second authentication request
    # while the token manager is obtaining a new access token.
    # Unfortunately, the TokenManager class  method does the refresh request synchronously,
    # so we get back the second access token here instead.
    # If we "fix" the TokenManager class to refresh asynchronously, we'll need to
    # change this test case to expect the first access token here.
    request = {'headers': {}}
    authenticator.authenticate(request)
    assert request['headers']['Authorization'] == 'Bearer ' + access_token2

    # Wait for the background refresh to finish.
    # No need to wait due to the synchronous logic in the TokenManager class mentioned above.
    # time.sleep(2)

    # Authenticate another request and verify that the second access token is used again.
    request = {'headers': {}}
    authenticator.authenticate(request)
    assert request['headers']['Authorization'] == 'Bearer ' + access_token2


@responses.activate
def test_request_token():
    (response, access_token) = get_mock_token_response(time.time(), 30)
    responses.add(responses.POST, MOCK_URL + MOCK_PATH, body=response, status=200)

    token_manager = MCSPV2TokenManager(
        apikey=MOCK_APIKEY,
        url=MOCK_URL,
        scope_collection_type=MOCK_SCOPE_COLLECTION_TYPE,
        scope_id=MOCK_SCOPE_ID,
        disable_ssl_verification=True,
    )
    token = token_manager.get_token()

    assert len(responses.calls) == 1
    assert (
        responses.calls[0].request.url
        == MOCK_URL
        + MOCK_PATH
        + '?includeBuiltinActions=false&includeCustomActions=false&'
        + 'includeRoles=true&prefixRolesWithDefinitionScope=false'
    )
    assert responses.calls[0].request.headers.get('User-Agent').startswith('ibm-python-sdk-core/mcspv2-authenticator')
    assert token == access_token


@responses.activate
def test_request_token_unsuccessful():
    response = """{
        "errorCode": "BXNIM0415E",
        "errorMessage": "Provided API key could not be found"
    }
    """
    responses.add(responses.POST, url=MOCK_URL + MOCK_PATH, body=response, status=400)

    token_manager = MCSPV2TokenManager(
        apikey="bad-api-key",
        url=MOCK_URL,
        scope_collection_type=MOCK_SCOPE_COLLECTION_TYPE,
        scope_id=MOCK_SCOPE_ID,
        disable_ssl_verification=True,
    )
    with pytest.raises(Exception):
        token_manager.request_token()

    assert len(responses.calls) == 1
    assert (
        responses.calls[0].request.url
        == MOCK_URL
        + MOCK_PATH
        + '?includeBuiltinActions=false&includeCustomActions=false&'
        + 'includeRoles=true&prefixRolesWithDefinitionScope=false'
    )
    assert responses.calls[0].response.text == response
